# coding=utf-8
import json
import time
from collections import namedtuple
from typing import Union, Dict, Optional
from functools import lru_cache, cached_property

from confluent_kafka.cimpl import KafkaException, Message as KafkaMessage

import scenarios.logging.logger_constants as log_const
from core.configs.global_constants import KAFKA_REPLY_TOPIC
from core.logging.logger_utils import log, UID_STR, MESSAGE_ID_STR
from smart_kit.names.message_names import ANSWER_TO_USER, RUN_APP, MESSAGE_TO_SKILL, SERVER_ACTION, CLOSE_APP

from core.message.from_message import SmartAppFromMessage
from core.model.base_user import BaseUser
from core.model.heapq.heapq_storage import HeapqKV
from core.mq.kafka.kafka_consumer import KafkaConsumer
from core.mq.kafka.kafka_publisher import KafkaPublisher
from core.utils.pickle_copy import pickle_deepcopy
from core.utils.stats_timer import StatsTimer
from core.basic_models.actions.command import Command
from smart_kit.compatibility.commands import combine_commands
from smart_kit.message.get_to_message import get_to_message
from smart_kit.message.smartapp_to_message import SmartAppToMessage
from smart_kit.names import message_names
from smart_kit.request.kafka_request import SmartKitKafkaRequest
from smart_kit.start_points.base_main_loop import BaseMainLoop
from core.monitoring.monitoring import monitoring


def _enrich_config_from_secret(kafka_config, secret_config):
    for key in kafka_config:
        if secret_config.get(key):
            kafka_config[key]["consumer"]["conf"].update(secret_config[key]["consumer"]["conf"])
            kafka_config[key]["publisher"]["conf"].update(secret_config[key]["publisher"]["conf"])
    return kafka_config


class MainLoop(BaseMainLoop):
    MAX_LOG_TIME = 20
    BAD_ANSWER_COMMAND = Command(message_names.ERROR, {"code": -1, "description": "Invalid Answer Message"})

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        log("%(class_name)s.__init__ started.", params={log_const.KEY_NAME: log_const.STARTUP_VALUE,
                                                        "class_name": self.__class__.__name__})
        try:
            kafka_config = _enrich_config_from_secret(
                self.settings["kafka"]["template-engine"], self.settings.get("secret_kafka", {})
            )

            consumers = {}
            publishers = {}
            log("%(class_name)s START CONSUMERS/PUBLISHERS CREATE",
                params={"class_name": self.__class__.__name__}, level="WARNING")
            kafka_config_copy = pickle_deepcopy(kafka_config)
            for key, config in kafka_config_copy.items():
                if config.get("consumer"):
                    consumers.update({key: KafkaConsumer(config)})
                if config.get("publisher"):
                    publishers.update({key: KafkaPublisher(config)})
            log("%(class_name)s FINISHED CONSUMERS/PUBLISHERS CREATE",
                params={"class_name": self.__class__.__name__}, level="WARNING")

            self.app_name = self.settings.app_name
            self.consumers = consumers
            for key in self.consumers:
                self.consumers[key].subscribe()
            self.publishers = publishers
            self.behaviors_timeouts_value_cls = namedtuple('behaviors_timeouts_value',
                                                           'db_uid, callback_id, mq_message, kafka_key')
            self.behaviors_timeouts = HeapqKV(value_to_key_func=lambda val: val.callback_id)
            log("%(class_name)s.__init__ completed.", params={log_const.KEY_NAME: log_const.STARTUP_VALUE,
                                                              "class_name": self.__class__.__name__})
        except:
            log("%(class_name)s.__init__ exception.", params={log_const.KEY_NAME: log_const.STARTUP_VALUE,
                                                              "class_name": self.__class__.__name__},
                level="ERROR", exc_info=True)
            raise

    def pre_handle(self):
        self.iterate_behavior_timeouts()

    def run(self):
        log("%(class_name)s.run started", params={log_const.KEY_NAME: log_const.STARTUP_VALUE,
                                                  "class_name": self.__class__.__name__})
        while self.is_work:
            self.pre_handle()
            for kafka_key in self.consumers:
                self.iterate(kafka_key)

            if self.health_check_server:
                with StatsTimer() as health_check_server_timer:
                    self.health_check_server.iterate()

                if health_check_server_timer.msecs >= self.MAX_LOG_TIME:
                    log("Health check iterate time: {} msecs\n".format(health_check_server_timer.msecs),
                        params={log_const.KEY_NAME: "slow_health_check",
                                "time_msecs": health_check_server_timer.msecs}, level="WARNING")

        log("Stopping Kafka handler", level="WARNING")
        for kafka_key in self.consumers:
            self.consumers[kafka_key].close()
            log("Kafka consumer connection is closed", level="WARNING")
            self.publishers[kafka_key].close()
            log("Kafka publisher connection is closed", level="WARNING")
        log("Kafka handler is stopped", level="WARNING")

    def _generate_answers(self, user, commands, message, **kwargs):
        topic_key = kwargs["topic_key"]
        kafka_key = kwargs["kafka_key"]
        answers = []
        commands = commands or []

        commands = combine_commands(commands, user)

        for command in commands:
            request = SmartKitKafkaRequest(id=None, items=command.request_data)
            request.update_empty_items({
                "kafka_key": kafka_key,
                "topic_key": topic_key,
                "topic": user.private_vars.get(KAFKA_REPLY_TOPIC) if command.name == ANSWER_TO_USER else None
            })

            to_message = get_to_message(command.name)
            answer = to_message(command=command, message=message, request=request,
                                masking_fields=self.masking_fields,
                                validators=self.to_msg_validators)
            if answer.validate():
                answers.append(answer)
            else:
                answers.append(SmartAppToMessage(self.BAD_ANSWER_COMMAND, message=message, request=request))

            monitoring.counter_outgoing(self.app_name, command.name, answer, user)

        return answers

    def _get_timeout_from_message(self, orig_message_raw: Dict, callback_id, headers):
        timeout_from_message = SmartAppFromMessage(orig_message_raw, headers=headers,
                                                   masking_fields=self.masking_fields,
                                                   validators=self.from_msg_validators)
        timeout_from_message.callback_id = callback_id
        return timeout_from_message

    def iterate_behavior_timeouts(self):
        now = time.time()
        while now > (self.behaviors_timeouts.get_head_key() or float("inf")):
            _, behavior_timeout_value = self.behaviors_timeouts.pop()
            db_uid, callback_id, mq_message, kafka_key = behavior_timeout_value
            try:
                save_tries = 0
                user_save_no_collisions = False
                user = None
                timeout_from_message = None
                answers = None
                while save_tries < self.user_save_collisions_tries and not user_save_no_collisions:
                    save_tries += 1

                    orig_message_raw = json.loads(mq_message.value())
                    orig_message_raw[SmartAppFromMessage.MESSAGE_NAME] = message_names.LOCAL_TIMEOUT

                    timeout_from_message = self._get_timeout_from_message(orig_message_raw, callback_id,
                                                                          headers=mq_message.headers())

                    user = self.load_user(db_uid, timeout_from_message)
                    commands = self.model.answer(timeout_from_message, user)
                    topic_key = self._get_topic_key(mq_message, kafka_key)
                    answers = self._generate_answers(user=user, commands=commands, message=timeout_from_message,
                                                     topic_key=topic_key,
                                                     kafka_key=kafka_key)

                    user_save_no_collisions = self.save_user(db_uid, user, mq_message)

                    if user and not user_save_no_collisions:
                        log("MainLoop.iterate_behavior_timeouts: save user got collision on uid %(uid)s db_version "
                            "%(db_version)s.",
                            user=user,
                            params={log_const.KEY_NAME: "ignite_collision",
                                    "db_uid": db_uid,
                                    "message_key": mq_message.key(),
                                    "kafka_key": kafka_key,
                                    "uid": user.id,
                                    "db_version": str(user.private_vars.get(user.USER_DB_VERSION))},
                            level="WARNING")

                        continue

                if not user_save_no_collisions:
                    log("MainLoop.iterate_behavior_timeouts: db_save collision all tries left on uid %(uid)s "
                        "db_version %(db_version)s.",
                        user=user,
                        params={log_const.KEY_NAME: "ignite_collision",
                                "db_uid": db_uid,
                                "message_key": mq_message.key(),
                                "message_partition": mq_message.partition(),
                                "kafka_key": kafka_key,
                                "uid": user.id,
                                "db_version": str(user.private_vars.get(user.USER_DB_VERSION))},
                        level="WARNING")

                    monitoring.counter_save_collision_tries_left(self.app_name)
                self.save_behavior_timeouts(user, mq_message, kafka_key)
                for answer in answers:
                    self._send_request(user, answer, mq_message)
            except:
                log("%(class_name)s error.", params={log_const.KEY_NAME: "error_handling_timeout",
                                                     "class_name": self.__class__.__name__,
                                                     log_const.REQUEST_VALUE: str(mq_message.value())},
                    level="ERROR", exc_info=True)

    def _get_topic_key(self, mq_message: KafkaMessage, kafka_key):
        topic_names_2_key = self._topic_names_2_key(kafka_key)
        return self.default_topic_key(kafka_key) or topic_names_2_key[mq_message.topic()]

    def process_message(self, mq_message: KafkaMessage, consumer, kafka_key, stats):
        topic_key = self._get_topic_key(mq_message, kafka_key)

        save_tries = 0
        user_save_no_collisions = False
        user = None
        db_uid = None
        message = None
        while save_tries < self.user_save_collisions_tries and not user_save_no_collisions:
            save_tries += 1
            message_value = json.loads(mq_message.value())
            message = SmartAppFromMessage(message_value,
                                          headers=mq_message.headers(),
                                          masking_fields=self.masking_fields,
                                          creation_time=consumer.get_msg_create_time(mq_message))

            if message.validate():
                waiting_message_time = 0
                if message.creation_time:
                    waiting_message_time = time.time() * 1000 - message.creation_time
                    stats += "Waiting message: {} msecs\n".format(waiting_message_time)

                stats += "Mid: {}\n".format(message.incremental_id)
                monitoring.sampling_mq_waiting_time(self.app_name, waiting_message_time / 1000)

                # check_message_key
                message_key = self._get_str_message_key(mq_message.key(), incremental_id=message.incremental_id,
                                                        uid=message.uid, user=user)
                valid_key = self._get_valid_message_key(message)
                if message_key != valid_key:
                    log(f"Failed to check Kafka message %(message_name)s key {message_key} != {valid_key}",
                        params={
                            log_const.KEY_NAME: "check_kafka_key_validation",
                            MESSAGE_ID_STR: message.incremental_id,
                            UID_STR: message.uid,
                            "kafka_key": kafka_key,
                            "message_key": message_key,
                            "message_name": message.type,
                        }, user=user,
                        level="WARNING")
                    if self.settings["template_settings"].get("wrong_key_resend", True):
                        dest_topic = mq_message.topic()
                        self.publishers[kafka_key].send_to_topic(mq_message.value(), valid_key, dest_topic,
                                                                 mq_message.headers())
                        log("Kafka message %(message_name)s with invalid Kafka message key %(message_key)s "
                            f"resend again with a valid key: '{valid_key}' to '{dest_topic}'",
                            params={
                                log_const.KEY_NAME: "kafka_message_key_recovery",
                                MESSAGE_ID_STR: message.incremental_id,
                                UID_STR: message.uid,
                                "kafka_key": kafka_key,
                                "message_key": message_key,
                                "message_name": message.type,
                            }, user=user,
                            level=self.settings["template_settings"].get("kafka_message_key_recovery_log_level",
                                                                         "DEBUG"))
                        break
                else:
                    log("INCOMING FROM TOPIC: %(topic)s partition %(message_partition)s HEADERS: %(headers)s DATA: "
                        "%(incoming_data)s",
                        params={log_const.KEY_NAME: "incoming_message",
                                "topic": mq_message.topic(),
                                "message_partition": mq_message.partition(),
                                "message_key": (mq_message.key() or b"").decode('utf-8', 'backslashreplace'),
                                "message_id": message.incremental_id,
                                "kafka_key": kafka_key,
                                "incoming_data": str(message.masked_value),
                                "length": len(message.as_str),
                                "headers": mq_message.headers(),
                                "waiting_message": waiting_message_time,
                                "surface": message.device.surface,
                                MESSAGE_ID_STR: message.incremental_id},
                        user=user)

                    db_uid = message.db_uid
                    with StatsTimer() as load_timer:
                        user = self.load_user(db_uid, message)
                    monitoring.sampling_load_time(self.app_name, load_timer.secs)
                    stats += "Loading time: {} msecs\n".format(load_timer.msecs)
                    if KAFKA_REPLY_TOPIC in message.headers and \
                            message.message_name in [RUN_APP, MESSAGE_TO_SKILL, SERVER_ACTION, CLOSE_APP]:
                        if user.private_vars.get(KAFKA_REPLY_TOPIC):
                            log("MainLoop.iterate: kafka_replyTopic collision",
                                params={log_const.KEY_NAME: "ignite_collision",
                                        "db_uid": db_uid,
                                        "message_key": mq_message.key(),
                                        "message_partition": mq_message.partition(),
                                        "kafka_key": kafka_key,
                                        "uid": user.id,
                                        "saved_topic": user.private_vars.get(KAFKA_REPLY_TOPIC),
                                        "current_topic": message.headers[KAFKA_REPLY_TOPIC]},
                                user=user, level="WARNING")
                        user.private_vars.set(KAFKA_REPLY_TOPIC, message.headers[KAFKA_REPLY_TOPIC])
                    with StatsTimer() as script_timer:
                        commands = self.model.answer(message, user)

                    answers = self._generate_answers(user=user, commands=commands, message=message,
                                                     topic_key=topic_key,
                                                     kafka_key=kafka_key)
                    monitoring.sampling_script_time(self.app_name, script_timer.secs)
                    stats += "Script time: {} msecs\n".format(script_timer.msecs)

                    with StatsTimer() as save_timer:
                        user_save_no_collisions = self.save_user(db_uid, user, message)

                    monitoring.sampling_save_time(self.app_name, save_timer.secs)
                    stats += "Saving time: {} msecs\n".format(save_timer.msecs)
                    if not user_save_no_collisions:
                        log("MainLoop.iterate: save user got collision on uid %(uid)s db_version %(db_version)s.",
                            user=user,
                            params={log_const.KEY_NAME: "ignite_collision",
                                    "db_uid": db_uid,
                                    "message_key": mq_message.key(),
                                    "message_partition": mq_message.partition(),
                                    "kafka_key": kafka_key,
                                    "uid": user.id,
                                    "db_version": str(user.private_vars.get(user.USER_DB_VERSION))},
                            level="WARNING")
                        continue

                    self.save_behavior_timeouts(user, mq_message, kafka_key)

                    if mq_message.headers() is None:
                        mq_message.set_headers([])

                    if answers:
                        for answer in answers:
                            with StatsTimer() as publish_timer:
                                self._send_request(user, answer, mq_message)
                            stats += "Publishing time: {} msecs".format(publish_timer.msecs)
                            log(stats, user=user)
            else:
                try:
                    data = message.masked_value
                except:
                    data = "<DATA FORMAT ERROR>"
                log(f"Message validation failed, skip message handling.",
                    params={log_const.KEY_NAME: "invalid_message",
                            "data": data,
                            "message_key": (mq_message.key() or b"").decode('utf-8', 'backslashreplace')},
                    level="ERROR")
                monitoring.counter_invalid_message(self.app_name)
        if user and not user_save_no_collisions:
            log("MainLoop.iterate: db_save collision all tries left on uid %(uid)s db_version %(db_version)s.",
                user=user,
                params={log_const.KEY_NAME: "ignite_collision",
                        "db_uid": db_uid,
                        "message_key": mq_message.key(),
                        "message_partition": mq_message.partition(),
                        "kafka_key": kafka_key,
                        "uid": user.id,
                        "db_version": str(user.private_vars.get(user.USER_DB_VERSION))},
                level="WARNING")
            self.postprocessor.postprocess(user, message)
            monitoring.counter_save_collision_tries_left(self.app_name)
        consumer.commit_offset(mq_message)

    def iterate(self, kafka_key):
        consumer = self.consumers[kafka_key]
        mq_message: Optional[KafkaMessage] = None
        message_value = None
        try:
            mq_message = None
            message_value = None
            with StatsTimer() as poll_timer:
                mq_message = consumer.poll()

            if mq_message:
                stats = "Polling time: {} msecs\n".format(poll_timer.msecs)
                message_value = mq_message.value()  # DRY!
                self.process_message(mq_message, consumer, kafka_key, stats)

        except KafkaException as kafka_exp:
            log("kafka error: %(kafka_exp)s. MESSAGE: {}.".format(message_value),
                params={log_const.KEY_NAME: log_const.STARTUP_VALUE,
                        "kafka_exp": str(kafka_exp),
                        log_const.REQUEST_VALUE: str(message_value)},
                level="ERROR", exc_info=True)
        except Exception:
            try:
                log("%(class_name)s iterate error. Kafka key %(kafka_key)s MESSAGE: {}.".format(message_value),
                    params={log_const.KEY_NAME: log_const.STARTUP_VALUE,
                            "kafka_key": kafka_key},
                    level="ERROR", exc_info=True)
                consumer.commit_offset(mq_message)
            except Exception:
                log("Error handling worker fail exception.",
                    level="ERROR", exc_info=True)

    def _get_valid_message_key(self, from_message: SmartAppFromMessage):
        return "_".join([i for i in [from_message.channel, from_message.sub, from_message.uid] if i])

    def _get_str_message_key(self, message_key: Union[str, bytes], incremental_id: int, uid: str, user: BaseUser):
        message_key = message_key or b""
        try:
            if isinstance(message_key, bytes):
                message_key = message_key.decode()
        except UnicodeDecodeError:
            log(f"MainLoop._get_str_message_key: Decode error of Kafka message key {message_key}",
                params={log_const.KEY_NAME: "check_kafka_key_error",
                        MESSAGE_ID_STR: incremental_id,
                        UID_STR: uid},
                user=user, level="ERROR")

        return message_key

    def _send_request(self, user: BaseUser, answer: SmartAppToMessage, mq_message: KafkaMessage):
        kafka_broker_settings = self.settings["template_settings"].get(
            "route_kafka_broker"
        ) or []

        request = answer.request

        for kb_setting in kafka_broker_settings:
            if (
                    kb_setting["from_channel"] == answer.incoming_message.channel
                    and kb_setting["to_topic"] == request.topic_key
            ):
                request.kafka_key = kb_setting["route_to_broker"]

        request_params = dict()
        request_params["publishers"] = self.publishers
        request_params["mq_message"] = mq_message
        request_params["payload"] = answer.value
        request_params["masked_value"] = answer.masked_value
        request.run(answer.value, request_params)
        self._log_request(user, request, answer, mq_message)

    def _log_request(self, user, request, answer, original_mq_message):
        log("OUTGOING TO TOPIC_KEY: %(topic_key)s DATA: %(data)s",
            params={log_const.KEY_NAME: "outgoing_message",
                    "topic": request.topic_key,
                    "headers": request._get_new_headers(original_mq_message),
                    "data": answer.masked_value,
                    "length": len(answer.value),
                    "message_key": (original_mq_message.key() or b"").decode('utf-8', 'backslashreplace')},
            user=user)

    @lru_cache()
    def _topic_names_2_key(self, kafka_key):
        topics = self.settings["kafka"]["template-engine"][kafka_key]["consumer"]["topics"]
        return {name: key for key, name in topics.items()}

    def default_topic_key(self, kafka_key):
        return self.settings["kafka"]["template-engine"][kafka_key].get("default_topic_key")

    @cached_property
    def masking_fields(self):
        return self.settings["template_settings"].get("masking_fields")

    def save_behavior_timeouts(self, user, mq_message: KafkaMessage, kafka_key):
        for i, (expire_time_us, callback_id) in enumerate(user.behaviors.get_behavior_timeouts()):
            # two behaviors can be created in one query, so we need add some salt to make theirs key unique
            unique_key = expire_time_us + i * 1e-5
            log("%(class_name)s: adding local_timeout on callback %(callback_id)s with timeout on %(unique_key)s",
                params={log_const.KEY_NAME: "adding_local_timeout",
                        "class_name": self.__class__.__name__,
                        log_const.BEHAVIOR_CALLBACK_ID_VALUE: callback_id,
                        "unique_key": unique_key})
            self.behaviors_timeouts.push(unique_key, self.behaviors_timeouts_value_cls._make(
                (user.message.db_uid, callback_id, mq_message, kafka_key)))

        for callback_id in user.behaviors.get_returned_callbacks():
            log("%(class_name)s: removing local_timeout on callback %(callback_id)s",
                params={log_const.KEY_NAME: "removing_local_timeout",
                        "class_name": self.__class__.__name__,
                        log_const.BEHAVIOR_CALLBACK_ID_VALUE: callback_id})
            self.behaviors_timeouts.remove(callback_id)

    def stop(self, signum, frame):
        self.is_work = False



import heapq


class HeapqKV:
    def __init__(self, value_to_key_func):
        self._heapq = []
        self._to_remove = set()
        self._value_to_key = value_to_key_func

    def value_to_key(self, value):
        return self._value_to_key(value)

    def push(self, key, value):
        heapq.heappush(self._heapq, (key, value))

    def get_head_key(self):
        while self._heapq:
            top = self._heapq[0]
            if self._check_key_valid(top[1]):
                return top[0]
            else:
                heapq.heappop(self._heapq)

    def pop(self):
        while self._heapq:
            top = self._heapq[0]
            if self._check_key_valid(top[1]):
                return heapq.heappop(self._heapq)
            else:
                heapq.heappop(self._heapq)

    def remove(self, key):
        self._to_remove.add(key)

    def _check_key_valid(self, value):
        key = self.value_to_key(value)
        if key in self._to_remove:
            self._to_remove.remove(key)
            return False
        return True



# coding: utf-8
from typing import Union, Dict, List, Any, Optional

from core.basic_models.actions.command import Command
from core.basic_models.actions.string_actions import StringAction
from core.text_preprocessing.base import BaseTextPreprocessingResult
from core.utils.utils import deep_update_dict
from scenarios.user.user_model import User
from smart_kit.request.kafka_request import SmartKitKafkaRequest

PUSH_NOTIFY = "PUSH_NOTIFY"


class PushAction(StringAction):
    """ Action для отправки пуш уведомлений в сервис пушей.
        Example:
        {
          "push_advertising_offer": {
            "type": "push",
            "surface": "COMPANION", // не обязательное, по дефолту "COMPANION", без шаблонной генерации
            "content": {
                "notificationHeader": "{% if day_time = 'morning' %}Время завтракать!{% else %}Хотите что нибудь заказать?{% endif %}",
                "fullText": "В нашем магазине большой ассортимент{% if day_time = 'evening' %}. Успей заказать!{% endif %}",
                "mobileAppParameters": {
                    "DeeplinkAndroid": "{[ deep_link_url }}",
                    "DeeplinkIos": "{{ deep_link_url }}",
                    "Logo": "{{ icon_url }}"
                }
            }
          }
        }
    """
    DEFAULT_EXTRA_HEADERS = {
        "request-id": "{{ uuid4() }}",
        "sender-id": "{{ uuid4() }}",
        "simple": "true"
    }

    DEFAULT_REQUEST_DATA = {
        "topic_key": "push",
        "kafka_key": "main",
        SmartKitKafkaRequest.KAFKA_EXTRA_HEADERS: DEFAULT_EXTRA_HEADERS
    }

    COMPANION = "COMPANION"
    EX_HEADERS_NAME = SmartKitKafkaRequest.KAFKA_EXTRA_HEADERS

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        self.surface = items.get("surface", self.COMPANION)
        items["nodes"] = items.get("content") or {}
        items["command"] = PUSH_NOTIFY
        super().__init__(items, id)
        self.default_request_data_template = self._get_template_tree(self.DEFAULT_REQUEST_DATA)
        self.request_data_template = self._get_template_tree(self.request_data) if self.request_data else None

    def _render_request_data(self, action_params):
        request_data = self._get_rendered_tree_recursive(self.default_request_data_template, action_params)
        if self.request_data:
            request_data_update = self._get_rendered_tree_recursive(self.request_data_template, action_params)
            request_data = deep_update_dict(request_data, request_data_update)
        return request_data

    def run(self, user: User, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        params = params or {}
        command_params = {
            "projectId": user.settings["template_settings"]["project_id"],
            "clientId": user.message.sub,
            "surface": self.surface,
            "content": self._generate_command_context(user, text_preprocessing_result, params),
        }
        requests_data = self._render_request_data(params)
        commands = [Command(self.command, command_params, self.id, request_type=self.request_type,
                            request_data=requests_data, need_payload_wrap=False, need_message_name=False)]
        return commands



# coding: utf-8
from functools import cached_property

from core.logging.logger_utils import log
from core.names import field

import scenarios.logging.logger_constants as log_const

from scenarios.scenario_descriptions.form_filling_scenario import FormFillingScenario
from smart_kit.system_answers.nothing_found_action import NothingFoundAction
from core.monitoring.monitoring import monitoring


class DialogueManager:
    NOTHING_FOUND_ACTION = "nothing_found_action"

    def __init__(self, scenario_descriptions, app_name, **kwargs):
        log(f"{self.__class__.__name__}.__init__ started.",
            params={log_const.KEY_NAME: log_const.STARTUP_VALUE})
        self.scenario_descriptions = scenario_descriptions
        self.scenarios = scenario_descriptions["scenarios"]
        self.scenario_keys = set(self.scenarios.get_keys())
        self.actions = scenario_descriptions["external_actions"]
        self.app_name = app_name
        log(f"{self.__class__.__name__}.__init__ finished.", params={log_const.KEY_NAME: log_const.STARTUP_VALUE})

    @cached_property
    def _nothing_found_action(self):
        return self.actions.get(self.NOTHING_FOUND_ACTION) or NothingFoundAction()

    def run(self, text_preprocessing_result, user):
        before_action = user.descriptions["external_actions"].get("before_action")
        if before_action:
            params = user.parametrizer.collect(text_preprocessing_result)
            before_action.run(user, text_preprocessing_result, params)
        scenarios_names = user.last_scenarios.scenarios_names
        scenario_key = user.message.payload[field.INTENT]
        if scenario_key in scenarios_names:
            scenario = self.scenarios[scenario_key]
            is_form_filling = isinstance(scenario, FormFillingScenario)
            if is_form_filling:
                if not scenario.text_fits(text_preprocessing_result, user):
                    params = user.parametrizer.collect(text_preprocessing_result)
                    if scenario.check_ask_again_requests(text_preprocessing_result, user, params):
                        reply = scenario.ask_again(text_preprocessing_result, user, params)
                        return reply, True
                    monitoring.counter_nothing_found(self.app_name, scenario_key, user)
                    return self._nothing_found_action.run(user, text_preprocessing_result), False
        return self.run_scenario(scenario_key, text_preprocessing_result, user), True

    def run_scenario(self, scen_id, text_preprocessing_result, user):
        initial_last_scenario = user.last_scenarios.last_scenario_name
        scenario = self.scenarios[scen_id]
        params = {log_const.KEY_NAME: log_const.CHOSEN_SCENARIO_VALUE,
                  log_const.CHOSEN_SCENARIO_VALUE: scen_id,
                  log_const.SCENARIO_DESCRIPTION_VALUE: scenario.scenario_description
                  }
        log(log_const.LAST_SCENARIO_MESSAGE, user, params)
        run_scenario_result = scenario.run(text_preprocessing_result, user)

        actual_last_scenario = user.last_scenarios.last_scenario_name
        if actual_last_scenario and actual_last_scenario != initial_last_scenario:
            monitoring.counter_scenario_change(self.app_name, actual_last_scenario, user)

        return run_scenario_result



from typing import Optional, Dict, Any, Union, List

from core.basic_models.actions.basic_actions import CommandAction, Action
from core.basic_models.actions.basic_actions import action_factory
from core.basic_models.actions.command import Command
from core.descriptions.smart_updatable_descriptions_items import SmartUpdatableDescriptionsItems
from core.model.base_user import BaseUser
from core.text_preprocessing.base import BaseTextPreprocessingResult


class ExternalActions(SmartUpdatableDescriptionsItems):
    def __init__(self, items: Dict[str, Any]):
        super(ExternalActions, self).__init__(action_factory, items)


class ExternalAction(CommandAction):
    version: Optional[int]

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super(ExternalAction, self).__init__(items, id)
        self._action_key: str = items["action"]

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        action: Action = user.descriptions["external_actions"][self._action_key]
        commands = action.run(user, text_preprocessing_result, params)
        return commands



# coding: utf-8
from core.model.lazy_items import LazyItems


class Fields(LazyItems):
    def __init__(self, items, descriptions, user, factory, lifetime):
        self._lifetime = lifetime
        super(Fields, self).__init__(items, descriptions, user, factory)

    def _build_factory(self, description, raw_data):
        return self._factory(description, raw_data or {}, self._user, self._lifetime)

    @property
    def values(self):
        result = dict()
        for key in self:
            item = self[key]
            result[item.description.id] = item.value
        return result



import os
import platform
import sys

from setuptools import find_packages, setup

import versioneer

with open("README.md", "r", encoding="utf-8") as file:
    long_description = file.read()


def tf_version():
    if sys.platform == "darwin" and platform.machine() == "arm64":
        return "tensorflow-macos==2.6.0"
    elif sys.platform == "darwin" and platform.machine() == "x86_64":
        return "tensorflow==2.6.0"
    elif sys.platform == "linux" and platform.machine() == "x86_64":
        return "tensorflow==2.6.0"
    elif sys.platform == "linux" and platform.machine() == "aarch64":
        return "tensorflow-aarch64==2.6.0"
    return "tensorflow==2.6.0"


setup(
    name=os.getenv("PACKAGE_NAME", default="smart_app_framework"),
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Salute-Developers",
    description="SmartApp Framework — это фреймворк, "
                "который позволяет создавать смартапы "
                "с поддержкой виртуальных ассистентов Салют.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=[]),
    include_package_data=True,
    install_requires=[
        "aiohttp==3.7.4",
        "aioredis==2.0.0",
        "boto==2.49.0",
        "confluent_kafka==1.9.2",
        "croniter==1.3.7",
        "dawg==0.8.0",
        "dill==0.3.3",
        "ics==0.6",
        "Jinja2==3.0.3",
        "keras==2.6.0",
        "nltk==3.5",
        "numpy==1.19.3",
        "objgraph==3.4.1",
        "prometheus-client==0.7.1",
        "psutil==5.8.0",
        "pyignite==0.5.2",
        "pymorphy2==0.8",
        "pymorphy2_dicts==2.4.393442.3710985",
        "python-dateutil==2.7.3",
        "python-json-logger==0.1.11",
        "PyYAML==5.3",
        "redis==4.1.4",
        "requests==2.22.0",
        "rusenttokenize==0.0.5",
        "scikit-learn==1.1.2",
        "setuptools==62.3.2",
        "tabulate==0.9.0",
        "tatsu==4.4.0",
        "grpcio==1.49.1",  # library for tensorflow
        "oauthlib==3.2.0",  # library for tensorflow
        "google-auth==2.12.0",  # library for tensorflow
        tf_version(),
        "timeout-decorator==0.4.1",
        "tqdm==4.64.1",
        "incremental==21.3.0",  # library for Twisted
        "Twisted==22.8.0",
        "freezegun==1.1.0",
        "protobuf==3.19.6"  # https://developers.google.com/protocol-buffers/docs/news/2022-05-06#python-updates
    ],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9"
    ]
)



# coding: utf-8
from typing import List, Dict, Any

from core.basic_models.actions.command import Command
from core.monitoring.monitoring import monitoring
from scenarios.user.user_model import User


class HandlerBase:
    TOPIC_KEY = "template_app"
    KAFKA_KEY = "main"

    def __init__(self, app_name: str):
        self.app_name = app_name

    def run(self, payload: Dict[str, Any], user: User) -> List[Command]:
        # отправка события о входящем сообщении в систему мониторинга
        monitoring.counter_incoming(self.app_name, user.message.message_name, self.__class__.__name__,
                                    user, app_info=user.message.app_info)
        return []



# coding: utf-8
from typing import Dict, Any

from scenarios.scenario_descriptions.form_filling_scenario import FormFillingScenario
from scenarios.scenario_descriptions.tree_scenario.tree_scenario_node import TreeScenarioNode
from core.model.factory import dict_factory
from core.monitoring.monitoring import monitoring
from core.logging.logger_utils import log
import scenarios.logging.logger_constants as log_const
from scenarios.scenario_models.history import Event, HistoryConstants

MAIN_INNER_FORM_DELIMETER = "__"


class TreeScenario(FormFillingScenario):
    def __init__(self, items, id):
        super(TreeScenario, self).__init__(items, id)
        self._start_node_key = items["start_node_key"]
        self._scenario_nodes = items["scenario_nodes"]
        self.scenario_nodes = self.build_scenario_nodes()

    @dict_factory(TreeScenarioNode)
    def build_scenario_nodes(self):
        return self._scenario_nodes

    def _field(self, form, text_preprocessing_result, user, params):
        current_node = self.get_current_node(user)
        internal_form = self._get_internal_form(form.forms, current_node.form_key)
        return self._find_field(internal_form, text_preprocessing_result, user, params)

    def _set_current_node_id(self, user, node_id):
        user.scenario_models[self.id].current_node = node_id

    def _add_loop_count(self, user, node_id):
        user.scenario_models[self.id].add_count(node_id)

    def _remove_current_node_id(self, user):
        user.scenario_models[self.id].current_node = None

    def get_current_node(self, user):
        current_node_id = user.scenario_models[self.id].current_node
        current_node = self.scenario_nodes[current_node_id]
        return current_node

    def get_next_node(self, user, node, text_preprocessing_result, params):
        available_node_keys = node.available_nodes
        for key in available_node_keys:
            node = self.scenario_nodes[key]
            log_params = {log_const.KEY_NAME: log_const.CHECKING_NODE_ID_VALUE,
                          log_const.CHECKING_NODE_ID_VALUE: node.id}
            log(log_const.CHECKING_NODE_ID_MESSAGE, user, log_params)
            requirement_result = node.requirement.check(text_preprocessing_result, user, params)
            if requirement_result:
                log_params = {log_const.KEY_NAME: log_const.CHOSEN_NODE_ID_VALUE,
                              log_const.CHOSEN_NODE_ID_VALUE: node.id}
                log(log_const.CHOSEN_NODE_ID_MESSAGE, user, log_params)
                self._add_loop_count(user, node.id)
                return node

    def _get_form(self, user):
        forms = user.forms
        form = forms[self.form_type]
        if not form or form.is_valid():
            form = forms.new(self.form_type)
            user.scenario_models[self.id].current_node = self._start_node_key
        else:
            form.touch()
        return form

    def _get_internal_form(self, forms, form_key):
        if form_key is not None:
            return forms[form_key] or forms.new(form_key)

    def get_fields_data(self, main_form, form_type):
        all_forms_fields = {}
        for node_id in self.scenario_nodes:
            inner_form_key = self.scenario_nodes[node_id].form_key
            form = self._get_internal_form(main_form.forms, inner_form_key)
            if form:
                main_inner_forms_key = "".join(
                    (self._clean_key(form_type), MAIN_INNER_FORM_DELIMETER, self._clean_key(inner_form_key)))
                form_field_data = super(TreeScenario, self).get_fields_data(form, main_inner_forms_key)
                all_forms_fields.update(form_field_data)
        return all_forms_fields

    @monitoring.got_histogram("scenario_time")
    def run(self, text_preprocessing_result, user, params: Dict[str, Any] = None):
        main_form = self._get_form(user)
        user.last_scenarios.add(self.id, text_preprocessing_result)
        user.preprocessing_messages_for_scenarios.add(text_preprocessing_result)
        current_node = self.get_current_node(user)
        new_node = current_node
        self._add_loop_count(user, new_node.id)
        internal_form, form = None, None
        fill_other = True
        on_filled_actions = []

        while new_node:
            current_node = new_node
            log("message id: {}, current scenario: {}, current node: {}".format(
                user.message.incremental_id, self.id, current_node.id), user, self._log_params())
            internal_form = self._get_internal_form(main_form.forms, current_node.form_key)
            data_extracted = {}

            validation_error_msg = None
            for field_key, field_descr in internal_form.description.fields.items():
                field = internal_form.fields[field_key]
                if field.available:
                    extracted = field_descr.filler.run(user, text_preprocessing_result, params)
                    if extracted is not None:
                        event = Event(type=HistoryConstants.types.FIELD_EVENT,
                                      scenario=self.root_id,
                                      node=current_node.id,
                                      content={HistoryConstants.content_fields.FIELD: field_key},
                                      result=HistoryConstants.event_results.FILLED)
                        user.history.add_event(event)
                        log_params = self._log_params()
                        log_params["message_id"] = user.message.incremental_id
                        log_params["field_key"] = field_key
                        log_params["extracted"] = extracted
                        log("message id: %(message_id)s, extracted data for field %(field_key)s: "
                            "%(extracted)s", user, log_params)
                    if extracted is not None and fill_other:
                        fill_other = fill_other and field_descr.fill_other
                        field_data = {field_key: extracted}
                        _validation_error_msg = self._validate_extracted_data(user, text_preprocessing_result,
                                                                               internal_form, field_data, params)
                        if _validation_error_msg:
                            # return only first validation message in form
                            validation_error_msg = validation_error_msg or _validation_error_msg
                        else:
                            data_extracted.update(field_data)
            on_filled_node_actions, is_break = self._fill_form(user, text_preprocessing_result,
                                                               internal_form, data_extracted)
            if is_break:
                return on_filled_node_actions
            on_filled_actions.extend(on_filled_node_actions)

            if validation_error_msg is not None:
                return validation_error_msg

            if internal_form.is_valid():
                if not current_node.available_nodes and not form:
                    # end node found
                    main_form.set_valid()
                    self._remove_current_node_id(user)
                    event = Event(
                        type=HistoryConstants.types.END_SCENARIO,
                        scenario=self.root_id,
                        node=current_node.id,
                        result=HistoryConstants.event_results.SUCCESS
                    )
                    user.history.add_event(event)
                    break

            elif not form:
                form = internal_form
                self._set_current_node_id(user, current_node.id)
            new_node = self.get_next_node(user, current_node, text_preprocessing_result, params)

        field = self._find_field(form, text_preprocessing_result,
                                                   user, params) if form else None

        reply_commands = on_filled_actions
        if field:
            event = Event(type=HistoryConstants.types.FIELD_EVENT,
                          scenario=self.root_id,
                          node=current_node.id,
                          content={HistoryConstants.content_fields.FIELD: field.description.id},
                          result=HistoryConstants.event_results.ASK_QUESTION)
            user.history.add_event(event)
        _command = self.get_reply(user, text_preprocessing_result, current_node.actions, field, main_form)
        reply_commands.extend(_command)

        if not reply_commands:
            reply_commands = self.get_no_commands_action(user, text_preprocessing_result)

        return reply_commands



# coding: utf-8


class PreprocessingMessagesDescription:
    def __init__(self, items):
        items = items or {}
        self.max_message_count = items["messages"]
        self.lifetime = items["lifetime"]



import json
import logging
from copy import copy
import jinja2
from distutils.util import strtobool

import core.logging.logger_constants as log_const
from core.logging.logger_utils import log
from core.monitoring.monitoring import monitoring

UNIFIED_TEMPLATE_TYPE_NAME = "unified_template"


def bool_loader(val):
    return bool(strtobool(val))


class UnifiedTemplate:
    loaders = {
        "str": str,
        "int": int,
        "float": float,
        "bool": bool_loader,
        "json": json.loads,
    }

    def __init__(self, input):
        self.input = input
        if isinstance(input, str):
            self.template = jinja2.Template(input)
            self.loader = UnifiedTemplate.loaders["str"]
            self.support_templates = dict()
        elif isinstance(input, dict):
            if input.get("type") != UNIFIED_TEMPLATE_TYPE_NAME:
                raise Exception("template must be string or dict with type='{}'".format(UNIFIED_TEMPLATE_TYPE_NAME))
            self.template = jinja2.Template(input["template"], extensions=input.get("extensions", ()))
            self.loader = UnifiedTemplate.loaders[input.get("loader", "str")]
            self.support_templates = {k: UnifiedTemplate(t) for k, t in input.get("support_templates", dict()).items()}
        else:
            raise Exception("template must be string or dict with type='{}'".format(UNIFIED_TEMPLATE_TYPE_NAME))
        self.is_logging_debug_mode = logging.getLogger(globals().get("__name__")).isEnabledFor(logging.getLevelName("DEBUG"))


    def render(self, *args, **kwargs):
        params_dict = dict(*args, **kwargs)
        try:
            result = self.silent_render(params_dict)
            if self.is_logging_debug_mode:
                log_params = dict()
                log_params[log_const.KEY_NAME] = log_const.TEMPLATE_TRACE_VALUE
                log_params["class_name"] = self.__class__.__name__
                log_params["rendered"] = str(result)
                log_params["raw_items"] = str(self.input)
                log_params["params"] = params_dict
                log(f"TRACING %(class_name)s. Result is %(rendered)s.", params=log_params, level="DEBUG")

        except Exception:
            log("Failed to render template: %(template)s with params %(params_dict_str)s",
                params={log_const.KEY_NAME: log_const.HANDLED_EXCEPTION_VALUE,
                        "template": str(self.input),
                        "params_dict_str": str(params_dict)},
                level="ERROR",
                exc_info=True)
            monitoring.got_counter("core_jinja_template_error")
            raise
        return result

    def silent_render(self, params_dict):
        if self.support_templates:
            changed_params_dict = copy(params_dict)
            for support_key, support_template in self.support_templates.items():
                changed_params_dict[support_key] = support_template.render(changed_params_dict)
        else:
            changed_params_dict = params_dict
        if changed_params_dict:
            result = self.template.render(changed_params_dict)
        else:
            result = self.template.render()
        if self.loader != str:
            result = self.loader(result)
        return result

    def __str__(self):
        return str(self.input)



import os
import requests
import json

'''
пример bundles.json

{
  "bundle_name": {
    "key.male_to_male": "Ты любил меня, а я тебя ненавидел",
    "key.male_to_female": "Ты любила меня, а я тебя ненавидел",
    "key.female_to_male": "Ты любил меня, а я тебя ненавидела",
    "key.female_to_female": "Ты любила меня, а я тебя ненавидела"
  }
}

пример scenarios.json с использовнаием bundle-ов

{
  "scenario_name": {
    "actions": [
      {
        "nodes": {
          "answer": [
            [
              "{{ gender_sensetive_text(bundle_name, key) }}"
            ]
          ]
        }
      }
    ]
  }
}

'''


class GetBundleCommand:
    def __init__(self, app_config):
        static_path = app_config.STATIC_PATH
        self.url = app_config.PPS_URL
        self.bundle_path = os.path.join(static_path, "references/bundles")
        self.file_path = os.path.join(self.bundle_path, "bundles.json")

    @staticmethod
    def get_default_items(text, key, bundle_name):
        return {
            bundle_name: {
                key: text
            }
        }

    @staticmethod
    def get_data(text, key, bundle_name):
        return json.dumps(
            {
                "payload": {
                    "items_to_pps": {
                        bundle_name: {
                            "type": "text",
                            "key": key,
                            "text": text,
                            "params": {}
                        }
                    }
                }
            }
        )

    def execute(self):
        print("insert text")
        text = input(">")
        print("insert key")
        key = input(">")
        print("insert bundle_name")
        bundle_name = input(">")
        if text and key and bundle_name:
            data = self.get_data(text, key, bundle_name)
            try:
                r = requests.post(url=self.url, data=data)
                if r.status_code == 200:
                    data = json.loads(r.content)
                    payload = data.get("payload")
                    items_from_pps = payload.get("items_from_pps")
                else:
                    items_from_pps = self.get_default_items(text, key, bundle_name)
            except:
                items_from_pps = self.get_default_items(text, key, bundle_name)
            if items_from_pps:
                if not os.path.exists(self.bundle_path):
                    os.makedirs(self.bundle_path)
                if os.path.exists(self.file_path):
                    with open(self.file_path, "r") as json_file:
                        file_data = json.load(json_file)
                        file_data.update(items_from_pps)
                else:
                    file_data = items_from_pps

                with open(self.file_path, "w") as json_file:
                    json.dump(file_data, json_file, ensure_ascii=False)



# coding: utf-8

from unittest import TestCase
from unittest.mock import Mock

from scenarios.scenario_models.field.fields import Fields
from smart_kit.utils.picklable_mock import PicklableMock


class TestFields(TestCase):

    def test_1(self):
        lifetime = 777
        descr1, descr2 = Mock(id="descr1"), Mock(id="descr2")
        items = {"descr1": PicklableMock(), "descr2": 2}
        descriptions = {"descr1": descr1, "descr2": descr2}

        factory = lambda descr, raw_data, user, lifetime: Mock(value=raw_data, lifetime=lifetime, description=descr)
        user = PicklableMock()
        fields = Fields(items, descriptions, user, factory, lifetime=lifetime)
        values = fields.values
        self.assertEqual(items, values)
        for field in fields:
            f = fields[field]
            self.assertEqual(lifetime, f.lifetime)



from typing import Sequence, List

from smart_kit.utils.cache import Cache


class BaseTextNormalizer:
    TEXT_PARAM_NAME = "text"
    PREPROCESS_METHOD = "preprocess"
    CLASSIFY_METHOD = "classify"
    NORMALIZE_METHOD = "normalize"
    CACHE = Cache

    def load_everything(self) -> None:
        raise NotImplementedError

    def with_cache(self, *args, **kwargs) -> 'BaseTextNormalizer':
        raise NotImplementedError

    def normalize_sequence(self, texts: Sequence, batch_size) -> List:
        raise NotImplementedError

    def __call__(self, text: str):
        raise NotImplementedError



# coding: utf-8


class BaseKafkaConsumer:
    def subscribe(self, topics=None):
        raise NotImplementedError

    def poll(self):
        raise NotImplementedError

    def consume(self, num_messages=1):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError



import unittest

from core.model.registered import registered_factories
from scenarios.scenario_models.field.field_filler_description import FieldFillerDescription, PreviousMessagesFiller
from scenarios.scenario_models.field.field_filler_description import field_filler_description, field_filler_factory
from smart_kit.utils.picklable_mock import PicklableMock


class MockFiller:
    def __init__(self, items=None):
        self.count = 0

    def extract(self, text_preprocessing_result, user, params):
        self.count += 1


class PreviousMessagesFillerTest(unittest.TestCase):
    def test_fill_1(self):
        registered_factories[FieldFillerDescription] = field_filler_factory
        field_filler_description["mock_filler"] = MockFiller
        expected = "first"
        items = {"filler": {"type": "mock_filler", "result": expected}}
        user = PicklableMock()
        user.preprocessing_messages_for_scenarios = PicklableMock()
        user.preprocessing_messages_for_scenarios.processed_items = [{}, {}, {}]
        filler = PreviousMessagesFiller(items)
        filler.extract(None, user)
        self.assertEqual(filler.filler.count, 4)

    def test_fill_2(self):
        registered_factories[FieldFillerDescription] = field_filler_factory
        field_filler_description["mock_filler"] = MockFiller
        expected = "first"
        items = {"filler": {"type": "mock_filler", "result": expected}, "count": 2}
        user = PicklableMock()
        user.preprocessing_messages_for_scenarios = PicklableMock()
        user.preprocessing_messages_for_scenarios.processed_items = [{}, {}, {}]
        filler = PreviousMessagesFiller(items)
        filler.extract(None, user)
        self.assertEqual(filler.filler.count, 2)



from unittest import TestCase

from core.utils.utils import convert_to_float
from core.utils.utils import get_int, get_number


class TestNumberConverter(TestCase):
    def test_1(self):
        sample = "10"
        expected = 10
        result = get_int(sample)
        self.assertEqual(expected, result)

    def test_2(self):
        sample = "-10"
        expected = -10
        result = get_int(sample)
        self.assertEqual(expected, result)

    def test_3(self):
        sample = "+10"
        expected = 10
        result = get_int(sample)
        self.assertEqual(expected, result)

    def test_4(self):
        sample = "123f"
        expected = None
        result = get_int(sample)
        self.assertEqual(expected, result)

    def test_5(self):
        sample = "99999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999"
        expected = 99999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999
        result = get_int(sample)
        self.assertEqual(expected, result)

    def test_6(self):
        sample = "10.5"
        expected = 10.5
        result = convert_to_float(sample)
        self.assertEqual(expected, result)

    def test_7(self):
        sample = "-10.5"
        expected = -10.5
        result = convert_to_float(sample)
        self.assertEqual(expected, result)

    def test_8(self):
        sample = "nan"
        expected = None
        result = convert_to_float(sample)
        self.assertEqual(expected, result)

    def test_9(self):
        sample = "nan"
        expected = None
        result = get_number(sample)
        self.assertEqual(expected, result)

    def test_10(self):
        sample = "-10"
        expected = -10
        result = get_number(sample)
        self.assertEqual(int, type(result))
        self.assertEqual(expected, result)

    def test_11(self):
        sample = "-10.008"
        expected = -10.008
        result = get_number(sample)
        self.assertEqual(float, type(result))
        self.assertEqual(expected, result)

    def test_12(self):
        sample = "один"
        result = convert_to_float(sample)
        self.assertIsNone(result)

    def test_13(self):
        sample = "один"
        result = get_int(sample)
        self.assertIsNone(result)

    def test_14(self):
        sample = "+79211234567"
        result = get_int(sample)
        self.assertIsNone(result)



from core.names import field


def get_callback_action_params(user):
    callback_id = user.message.callback_id
    action_params = user.behaviors.get_callback_action_params(callback_id) or {}
    return action_params


def set_debug_info(app_name, action_params, error):
    saved_debug_info = action_params.setdefault(field.DEBUG_INFO, {})
    dp_debug_info = saved_debug_info.setdefault(app_name, [])
    dp_debug_info.append({"error": error, "status": "{}_error".format(app_name)})




import os
from collections import OrderedDict
from typing import Callable, Any, Dict

import tensorflow as tf
from keras.utils.generic_utils import CustomObjectScope

import scenarios.logging.logger_constants as scenarios_log_const
from core.basic_models.classifiers.basic_classifiers import SkipClassifier, SciKitClassifier, ExternalClassifier, \
    SUPPORTED_CLASSIFIERS_TYPES
from core.basic_models.classifiers.classifiers_constants import REQUIRED_CONFIG_PARAMS
from core.logging.logger_utils import log
from core.repositories.base_repository import BaseRepository
from core.repositories.dill_repository import DillRepository
from core.repositories.folder_repository import FolderRepository
from core.text_preprocessing.preprocessing_result import TextPreprocessingResult

CLASSIFIER_TYPES_MAP = OrderedDict({"scikit": SciKitClassifier, "skip": SkipClassifier, "external": ExternalClassifier})


def classifiers_initial_launch(classifiers: Dict[str, Any]) -> None:
    # external классификаторы должны запускаться последними
    type_order = [_type for _type in CLASSIFIER_TYPES_MAP]
    sorted_cls_by_type_order = OrderedDict(sorted(classifiers.items(), key=lambda i: type_order.index(i[1]["type"])))

    text_preprocessing_result = TextPreprocessingResult({})
    for cls_name, cls_settings in sorted_cls_by_type_order.items():
        cls = CLASSIFIER_TYPES_MAP[cls_settings["type"]](cls_settings)
        cls.initial_launch(text_preprocessing_result, sorted_cls_by_type_order)
        sorted_cls_by_type_order[cls_name] = cls


class ClassifierRepository(BaseRepository):

    def __init__(self, description_path: str, data_path: str, loader: Callable, source: str, *args, **kwargs) -> None:
        super(ClassifierRepository, self).__init__(source=source, *args, **kwargs)
        self._description_path = description_path
        self._data_path = data_path
        self._required_classifier_config_params = REQUIRED_CONFIG_PARAMS
        self._supported_classifiers_types = SUPPORTED_CLASSIFIERS_TYPES
        self._folder_repository = FolderRepository(
            self._description_path, loader, source, *args, **kwargs) if self._check_paths_existence() else None

    def _subfolder_data_path(self, filename: str) -> str:
        return os.path.join(self._data_path, filename)

    def _check_paths_existence(self) -> bool:
        res = False
        # Проверяем что существуют обе директории: с конфигами классификаторов и чекпоинтами моделей,
        # также проверяем что эти обе директории не пустые
        if (os.path.exists(self._description_path) and len(os.listdir(self._description_path))
                and os.path.exists(self._data_path) and len(os.listdir(self._data_path))):
            res = True
        return res

    def _check_classifier_config(self, classifier_key: str, classifier_params: Dict[str, Any]) -> None:
        for req_param in self._required_classifier_config_params:
            try:
                classifier_params[req_param]
            except KeyError:
                raise Exception(f"Missing field: '{req_param}' for classifier {classifier_key} in classifiers.json")

    def load(self) -> None:
        if not self._folder_repository:
            return None

        self._folder_repository.load()
        classifiers_dict = self._folder_repository.data

        gpu_available = tf.test.is_gpu_available()
        repository = None

        for classifier_key in classifiers_dict:
            classifier_params = classifiers_dict[classifier_key]
            self._check_classifier_config(classifier_key, classifier_params)

            # Не грузить модель если она для gpu, но доступных gpu нет
            if classifier_params.get("is_gpu") and gpu_available is False:
                continue

            classifier_type = classifier_params["type"]
            if classifier_type not in self._supported_classifiers_types:
                log(
                    message=f"classifier_repository.load: Invalid classifier type for classifier "
                            f"%({scenarios_log_const.CLASSIFIER_VALUE})s",
                    params={scenarios_log_const.KEY_NAME: scenarios_log_const.STARTUP_VALUE,
                            scenarios_log_const.CLASSIFIER_VALUE: classifier_key},
                    level='WARN'
                )
                classifiers_dict[classifier_key] = SkipClassifier.get_nothing()
                continue

            # Нечего загружать тк в конфигурациях этих классификаторов модель не предусматривается
            if classifier_type in ["skip", "external"]:
                continue

            custom_layers = classifier_params.get("custom_layers", dict())
            if custom_layers:
                # Загрузка кастомных слоев, нужных для keras моделей
                try:
                    for layer in custom_layers:
                        layer_repository = DillRepository(self._subfolder_data_path(layer["path"]), self.source)
                        layer_repository.load()
                        custom_layers[layer["layer_name"]] = layer_repository.data
                except FileNotFoundError:
                    log(
                        message=f"classifier_repository.load: Failed to load custom layers for "
                        f"classifier %({scenarios_log_const.CLASSIFIER_VALUE})s, file not found",
                        params={scenarios_log_const.KEY_NAME: scenarios_log_const.STARTUP_VALUE,
                                scenarios_log_const.CLASSIFIER_VALUE: classifier_key},
                        level="WARN"
                    )
                    classifiers_dict[classifier_key] = SkipClassifier.get_nothing()
                    continue

            if classifier_type in ["scikit"]:
                repository = DillRepository(self._subfolder_data_path(classifier_params["path"]), self.source)

            log(
                message=f"classifier_repository.load: loading %({scenarios_log_const.CLASSIFIER_VALUE})s classifier",
                params={scenarios_log_const.KEY_NAME: scenarios_log_const.STARTUP_VALUE,
                        scenarios_log_const.CLASSIFIER_VALUE: classifier_key}
            )

            try:
                with CustomObjectScope(custom_layers):
                    repository.load()
                classifier_params["classifier"] = repository.data
            except FileNotFoundError:
                log(
                    message=f"classifier_repository.load: Failed to load classifier "
                            f"%({scenarios_log_const.CLASSIFIER_VALUE})s, file not found",
                    params={scenarios_log_const.KEY_NAME: scenarios_log_const.STARTUP_VALUE,
                            scenarios_log_const.CLASSIFIER_VALUE: classifier_key},
                    level="WARN"
                )
                classifiers_dict[classifier_key] = SkipClassifier.get_nothing()

        super(ClassifierRepository, self).fill(classifiers_dict)
        classifiers_initial_launch(classifiers_dict)

    def save(self, save_parameters: Any) -> None:
        pass



from core.descriptions.descriptions_items import DescriptionsItems


class SmartUpdatableDescriptionsItems(DescriptionsItems):

    def update_data(self, items):
        for item_id in items.keys():
            existed_item = self._items.get(item_id)
            if existed_item:
                if existed_item.version != items[item_id].get("version", -1) or \
                        items[item_id].get("force_update"):
                    del self._items[item_id]
        redundant = set(self._items.keys()) - set(items.keys())
        for key in redundant:
            del self._items[key]
        self._raw_items = items



import base64
import codecs
import json
import functools
import jinja2
import datetime
import pymorphy2
import uuid
import ics
import hashlib
import random
from pymorphy2.shapes import restore_capitalization
from core.unified_template.num2ordinal import Num2Ordinal
from core.unified_template.num2text import Num2Text
from urllib.parse import quote, unquote
from core.unified_template.currency2text import Money2Text, CURRENCY_CODE_2_NAME_MAP
from core.unified_template.pan2service import pan2service

morph = pymorphy2.MorphAnalyzer()
num2text = Num2Text()
num2ord = Num2Ordinal(num2text=num2text)
money2text = Money2Text(morph=morph)


def inflect(morph, text, required_grammemes):
    inflected = []
    for token in text.split():
        inflected_Parse_obj = morph.parse(token)[0].inflect(set(required_grammemes))
        inflected.append(restore_capitalization(inflected_Parse_obj.word, token) if inflected_Parse_obj else token)
    return " ".join(inflected)


def add_timedelta(date_time, *args, **kwargs):
    return date_time + datetime.timedelta(*args, **kwargs)


def okko_sec2text(s):
    h = int(s / 3600)
    m = int((s - h * 3600) / 60)
    return "{} ч {} мин".format(h, m) if h else "{} мин".format(m)


def generate_ics(date):
    calendar = ics.Calendar()
    event = ics.Event(begin=date)
    calendar.events.add(event)
    return str(calendar)


def timestamp_to_datetime(timestamp, utc_offset=0):
    return datetime.datetime.fromtimestamp(timestamp, datetime.timezone.utc) + datetime.timedelta(seconds=utc_offset)


def wrap_datetime(day, month, year, hour, minute, utc_offset=0):
    datetime_ = datetime.datetime(
        day=day, month=month, year=year, hour=hour, minute=minute, tzinfo=datetime.timezone.utc
    )
    return datetime_ + datetime.timedelta(seconds=utc_offset)


@jinja2.environmentfilter
def co_sort(
        environment, value, reverse=False, attribute=None, order=None,
):
    # custom order sort

    order = order or []

    def index(item):
        try:
            return order.index(item)
        except ValueError:
            return int(functools.reduce(lambda x, y: str(x) + str(y), map(ord, item)))

    key_func = jinja2.filters.make_attrgetter(
        environment, attribute,
        postprocess=index
    )
    return sorted(value, key=key_func, reverse=reverse)


def phone_number_ru(number):
    """Convert a 10 character string into +7(xxx) xxx-xx-xx."""
    first = number[0:3]
    second = number[3:6]
    third = number[6:8]
    fourth = number[8:10]
    return f"+7 ({first}) {second}-{third}-{fourth}"


def timestamp_to_date(raw_datetime):
    ru_month_names = {1: 'января',
                      2: 'февраля',
                      3: 'марта',
                      4: 'апреля',
                      5: 'мая',
                      6: 'июня',
                      7: 'июля',
                      8: 'августа',
                      9: 'сентября',
                      10: 'октября',
                      11: 'ноября',
                      12: 'декабря'}

    client_datetime = int(str(raw_datetime)[:10])

    stamp_datetime = datetime.datetime.fromtimestamp(client_datetime)
    normal_day, month_number = stamp_datetime.day, stamp_datetime.month
    show_date = f'{normal_day} {ru_month_names[month_number]}'

    return show_date

# functions
jinja2.defaults.DEFAULT_NAMESPACE["datetime_today"] = datetime.datetime.today
jinja2.defaults.DEFAULT_NAMESPACE["now"] = datetime.datetime.utcnow
jinja2.defaults.DEFAULT_NAMESPACE["uuid4"] = lambda: str(uuid.uuid4())
jinja2.defaults.DEFAULT_NAMESPACE["timestamp_to_datetime"] = timestamp_to_datetime
jinja2.defaults.DEFAULT_NAMESPACE["datetime"] = wrap_datetime
jinja2.defaults.DEFAULT_NAMESPACE["timedelta"] = datetime.timedelta
jinja2.defaults.DEFAULT_NAMESPACE["generate_ics"] = generate_ics

# ordinals
jinja2.filters.FILTERS["make_agree_with_number"] = lambda word, n: morph.parse(word)[0].make_agree_with_number(n).word
jinja2.filters.FILTERS["num2text"] = num2text
jinja2.filters.FILTERS["num2ord"] = num2ord
jinja2.filters.FILTERS["num2text_in_text"] = num2text.replace_everything_in_text
jinja2.filters.FILTERS["num2ord_in_text"] = num2ord.replace_everything_in_text
jinja2.filters.FILTERS["money2text"] = lambda amount, currency: money2text(amount, currency)
jinja2.filters.FILTERS["currency2name"] = lambda currency: CURRENCY_CODE_2_NAME_MAP.get(f"{currency}".upper(), currency)

# time
jinja2.filters.FILTERS["add_timedelta"] = add_timedelta
jinja2.filters.FILTERS["timestamp_to_time"] = lambda ts: datetime.datetime.fromtimestamp(ts)
jinja2.filters.FILTERS["strftime"] = lambda datetime, fmt: datetime.strftime(fmt)
jinja2.filters.FILTERS["timestamp_to_date"] = timestamp_to_date

# numbers
jinja2.filters.FILTERS["comma_float"] = lambda x: float(x.replace(",", "."))
jinja2.filters.FILTERS["careful_int"] = lambda x: int(x) if x.is_integer() else x
jinja2.filters.FILTERS["reduce"] = lambda num, reduce_num: int(num) - int(reduce_num)
jinja2.filters.FILTERS["increase"] = lambda num, increase_num: int(num) + int(increase_num)

# str
jinja2.filters.FILTERS["strip"] = str.strip
jinja2.filters.FILTERS["rstrip"] = str.rstrip
jinja2.filters.FILTERS["lstrip"] = str.lstrip
jinja2.filters.FILTERS["title"] = str.title
jinja2.filters.FILTERS["inflect"] = lambda text, required_grammemes: inflect(morph, text, required_grammemes)
jinja2.filters.FILTERS["url_quote"] = quote
jinja2.filters.FILTERS["url_unquote"] = unquote
jinja2.filters.FILTERS["hash_md5"] = lambda s: hashlib.md5(s.encode()).hexdigest()

# lists and sets
jinja2.filters.FILTERS["intersect"] = lambda x, y: set(x) & set(y)
jinja2.filters.FILTERS["difference"] = lambda x, y: set(x) - set(y)
jinja2.filters.FILTERS["co_sort"] = co_sort

# other
jinja2.filters.FILTERS["tojson"] = lambda x: json.dumps(x, ensure_ascii=False)
jinja2.filters.FILTERS["fromjson"] = json.loads
jinja2.filters.FILTERS["any"] = any
jinja2.filters.FILTERS["all"] = all
jinja2.filters.FILTERS["pan2service"] = pan2service
jinja2.filters.FILTERS["shuffle"] = lambda l: random.sample(l, len(l))

# absolutely custom
jinja2.filters.FILTERS["okko_sec2text"] = okko_sec2text

# Base64 encoding-decoding
# Additional padding in decode functions added to prevent 'Invalid padding error' for incorrectly-padded base64 strings
# len(input_string) must be divisible by 4, if not it should be padded with '=' char to match the requirement
# '===' is added to cover all cases, if string ends with '=' it is truncated to the nearest multiplier of 4 (internal impl)
jinja2.filters.FILTERS["b64decode"] = lambda a: base64.b64decode(a + '===')
jinja2.filters.FILTERS["b64encode"] = base64.b64encode
jinja2.filters.FILTERS["urlsafe_b64decode"] = lambda a: base64.urlsafe_b64decode(a + '===')
jinja2.filters.FILTERS["urlsafe_b64encode"] = base64.urlsafe_b64encode
jinja2.filters.FILTERS["tob64urlsafe"] = lambda x: base64.urlsafe_b64encode(bytes(x, 'UTF-8')).decode('utf-8')

# encoding
jinja2.filters.FILTERS["codecs_decode"] = lambda b, encoding='utf-8': codecs.decode(b, encoding=encoding)
jinja2.filters.FILTERS["codecs_encode"] = lambda s, encoding='utf-8': codecs.decode(s, encoding=encoding)

#localization
jinja2.filters.FILTERS["phone_number_ru"] = phone_number_ru



import time
from typing import Dict, Any, Optional


class Variables:
    DEFAULT_TTL = 86400

    def __init__(self, items, user, savable: bool = True):
        self._savable = savable
        self._storage: Dict[str, Any] = items or {}

    @property
    def raw(self) -> Optional[Dict[str, Any]]:
        if self._savable:
            return self._storage
        return None

    @property
    def values(self) -> Dict[str, Any]:
        self.expire()
        result = {}
        for key in self._storage:
            value, _ = self._storage[key]
            result[key] = value
        return result

    def set(self, key, value, ttl=None) -> None:
        ttl = ttl if ttl is not None else self.DEFAULT_TTL
        self._storage[key] = value, time.time() + ttl

    def update(self, key, value, ttl=None) -> None:
        _, old_ttl = self._storage[key]
        ttl = ttl or old_ttl
        self.set(key, value, ttl)

    def get(self, key, default=None):
        value, expire_time = self._storage.get(key, (default, time.time() + self.DEFAULT_TTL))
        if expire_time <= time.time():
            value = default
        return value

    def expire(self) -> None:
        for key in list(self._storage):
            _, expire_time = self._storage[key]
            if expire_time <= time.time():
                self.delete(key)

    def delete(self, key) -> None:
        del self._storage[key]

    def clear(self) -> None:
        self._storage.clear()



# coding: utf-8
from typing import Dict, Any


class Model:
    @property
    def fields(self):
        return []

    def __init__(self, values, user):
        values = values or {}

        for field in self.fields:
            value = values.get(field.name)
            description = field.description
            args = field.args
            if description is not None:
                obj = field.model(value, description, user, *args)
            else:
                obj = field.model(value, user, *args)
            setattr(self, field.name, obj)

    def get_field(self, name):
        return getattr(self, name)

    @property
    def raw(self) -> Dict[str, Any]:
        result = {}
        for field in self.fields:
            value = getattr(self, field.name)
            raw = value.raw
            if raw is not None:
                result[field.name] = raw
        return result



from typing import List, Any, Dict

from core.basic_models.actions.command import Command
from core.logging.logger_utils import log
from core.names.field import MESSAGE
from core.text_preprocessing.preprocessing_result import TextPreprocessingResult
from scenarios.user.user_model import User

from smart_kit.handlers.handler_base import HandlerBase
from smart_kit.names.field import PROFILE_DATA, STATUS_CODE, CODE, GEO


class HandlerTakeProfileData(HandlerBase):
    SUCCESS_CODE = 1

    def run(self, payload: Dict[str, Any], user: User) -> List[Command]:
        commands = super().run(payload, user)
        log(f"{self.__class__.__name__} started", user)

        if payload.get(STATUS_CODE, {}).get(CODE) == self.SUCCESS_CODE:
            commands.extend(user.behaviors.success(user.message.callback_id))
            user.variables.set("smart_geo", payload.get(PROFILE_DATA, {}).get(GEO))
        else:
            commands.extend(user.behaviors.fail(user.message.callback_id))
        return commands



# coding: utf-8
import random
from typing import Union, Dict, List, Any, Optional

import core.logging.logger_constants as log_const
from core.basic_models.actions.command import Command
from core.basic_models.requirement.basic_requirements import Requirement
from core.configs.global_constants import KAFKA
from core.logging.logger_utils import log
from core.model.base_user import BaseUser
from core.model.factory import build_factory, factory, list_factory
from core.model.registered import Registered
from core.text_preprocessing.base import BaseTextPreprocessingResult

actions = Registered()
action_factory = build_factory(actions)


class Action:
    """ Базовый класс для действий в сценариях, хранящий информацию о сценарии: id, version.

    Запуск действия обеспечивается методом run. В случае ошибки выполнения вызывается метод on_run_error. Атрибут id
    используется в качестве идентификатора действия, по которому можно вызвать действие. Атрибут version используется
    для отслеживания версии объекта, например, при обновлении SmartUpdatableDescriptionsItems.
    """
    version: Optional[int]
    id: Optional[str]

    def __init__(self, items: Optional[Dict[str, Any]] = None, id: Optional[str] = None):
        items = items or {}
        self.id = id
        self.version = items.get("version", -1)

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        return []

    def on_run_error(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser):
        log("exc_handler: Action failed to run. Return None. MESSAGE: %(masked_message)s.", user,
            {log_const.KEY_NAME: log_const.HANDLED_EXCEPTION_VALUE,
             "masked_message": user.message.masked_value},
            level="ERROR", exc_info=True)
        return None


class CommandAction(Action):
    """ Базовый класс для запроса к другому сервису, хранящий информацию о типе, методе и данных запроса.

    """
    DEFAULT_REQUEST_TYPE = KAFKA
    version: Optional[int]
    command: str
    request_type: Optional[str]
    request_data: Optional[Dict[str, str]]

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super().__init__(items, id)
        items = items or {}
        self.command = items.get("command")
        self.request_type = items.get("request_type") or self.DEFAULT_REQUEST_TYPE
        self.request_data = items.get("request_data")


class DoingNothingAction(CommandAction):
    version: Optional[int]
    command: str
    nodes: Dict[str, str]

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super().__init__(items, id)
        self.nodes = items.get("nodes") or {}

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        commands = super().run(user, text_preprocessing_result, params)
        commands.append(Command(self.command, self.nodes, self.id, request_type=self.request_type,
                                request_data=self.request_data))
        return commands


class RequirementAction(Action):
    version: Optional[int]
    requirement: Requirement
    internal_item: Action

    FIELD_KEY = "action"

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super().__init__(items, id)
        self._requirement: str = items["requirement"]
        # can be used not only with actions but with every entity which implements Action interface
        # to not change statics "item" key is added
        self._item: str = items[self.FIELD_KEY]

        self.requirement = self.build_requirement()
        self.internal_item = self.build_internal_item()

    @factory(Requirement)
    def build_requirement(self) -> str:
        return self._requirement

    @factory(Action)
    def build_internal_item(self) -> str:
        return self._item

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        commands = super().run(user, text_preprocessing_result, params)
        if self.requirement.check(text_preprocessing_result, user, params):
            commands.extend(self.internal_item.run(user, text_preprocessing_result, params))
        return commands


class ChoiceAction(Action):
    version: Optional[int]
    items: List[RequirementAction]
    else_item: Optional[Action]

    FIELD_REQUIREMENT_KEY = "requirement_actions"
    FIELD_ELSE_KEY = "else_action"

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super().__init__(items, id)
        self._requirement_items: List[str] = items[self.FIELD_REQUIREMENT_KEY]
        self._else_item: Optional[str] = items.get(self.FIELD_ELSE_KEY)

        self.items = self.build_items()

        if self._else_item:
            self.else_item = self.build_else_item()
        else:
            self.else_item = None

    @list_factory(RequirementAction)
    def build_items(self) -> List[str]:
        return self._requirement_items

    @factory(Action)
    def build_else_item(self) -> Optional[str]:
        return self._else_item

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        commands = super().run(user, text_preprocessing_result, params)
        choice_is_made = False
        for item in self.items:
            checked = item.requirement.check(text_preprocessing_result, user, params)
            if checked:
                commands.extend(item.internal_item.run(user, text_preprocessing_result, params))
                choice_is_made = True
                break
        if not choice_is_made and self._else_item:
            commands.extend(self.else_item.run(user, text_preprocessing_result, params))
        return commands


class ElseAction(Action):
    version: Optional[int]
    requirement: Requirement
    item: Action
    else_item: Optional[Action]

    FIELD_ITEM_KEY = "action"
    FIELD_ELSE_KEY = "else_action"

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super().__init__(items, id)
        self._requirement: str = items["requirement"]
        self._item: str = items[self.FIELD_ITEM_KEY]
        self._else_item: Optional[str] = items.get(self.FIELD_ELSE_KEY)

        self.requirement = self.build_requirement()
        self.item = self.build_item()
        if self._else_item:
            self.else_item = self.build_else_item()
        else:
            self.else_item = None

    @factory(Requirement)
    def build_requirement(self) -> str:
        return self._requirement

    @factory(Action)
    def build_item(self) -> str:
        return self._item

    @factory(Action)
    def build_else_item(self) -> Optional[str]:
        return self._else_item

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Optional[Dict[str, Union[str, float, int]]]] = None) -> List[Command]:
        commands = super().run(user, text_preprocessing_result, params)
        if self.requirement.check(text_preprocessing_result, user, params):
            commands.extend(self.item.run(user, text_preprocessing_result, params))
        elif self._else_item:
            commands.extend(self.else_item.run(user, text_preprocessing_result, params))
        return commands


class ActionOfActions(Action):
    version: Optional[int]
    actions: List[Action]

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super().__init__(items, id)
        self._actions: List[str] = items.get("actions") or []
        self.actions = self.build_actions()

    @list_factory(Action)
    def build_actions(self) -> List[Action]:
        return self._actions

class CompositeAction(ActionOfActions):
    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        commands = super().run(user, text_preprocessing_result, params)
        for action in self.actions:
            commands.extend(action.run(user, text_preprocessing_result, params))
        return commands


class NonRepeatingAction(ActionOfActions):
    last_action_ids_storage: str

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super().__init__(items, id)
        self._actions_count = len(items["actions"])
        self._last_action_ids_storage = items["last_action_ids_storage"]

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        commands = super().run(user, text_preprocessing_result, params)
        last_ids = user.last_action_ids[self._last_action_ids_storage]
        all_indexes = list(range(self._actions_count))
        max_last_ids_count = self._actions_count - 1
        # get last_actions_ids slice with max_len of max_last_ids_count
        last_actions_ids = last_ids.get_list()[-max_last_ids_count:]
        available_indexes = list(set(all_indexes) - set(last_actions_ids))
        action_index = random.choice(available_indexes)
        action = self.actions[action_index]
        last_ids.add(action_index)
        commands.extend(action.run(user, text_preprocessing_result, params))
        return commands


class RandomAction(Action):
    actions: List[Action]

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super().__init__(items, id)
        self._raw_actions: List[str] = items["actions"]
        self.actions = self.build_actions()

    @list_factory(Action)
    def build_actions(self) -> List[Action]:
        return self._raw_actions

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        commands = super().run(user, text_preprocessing_result, params)
        pos = random.randint(0, len(self._raw_actions) - 1)
        action = self.actions[pos]
        commands.extend(action.run(user, text_preprocessing_result, params=params))
        return commands



# coding: utf-8
import unittest

from core.basic_models.actions.basic_actions import Action, actions, action_factory
from core.model.registered import registered_factories
from scenarios.behaviors.behavior_description import BehaviorDescription
from smart_kit.utils.picklable_mock import PicklableMock


class MockAction(Action):
    def run(self, user, text_preprocessing_result, params=None):
        return []


class TestAction:
    def __init__(self):
        self.id = '123-456-789'
        self.version = 21  # какая-то версия

    def run(self, a, b, c):
        return 123   # некий результат


class BehaviorDescriptionTest(unittest.TestCase):

    def setUp(self):
        self.test_items = {"success_action": [TestAction()]}
        self.test_user = PicklableMock()
        self.test_user.settings = {"template_settings": {}}
        registered_factories[Action] = action_factory
        actions[None] = MockAction

    def test_create(self):
        behavior_description = BehaviorDescription({"success_action": None, "fail_action": None, "timeout": 10})
        self.assertEqual(behavior_description.timeout(self.test_user), 10)
        self.assertIsInstance(behavior_description.success_action, Action)
        self.assertIsInstance(behavior_description.fail_action, Action)

    def test_behavior_description_init(self):
        obj = BehaviorDescription(self.test_items)
        self.assertTrue(obj.timeout(self.test_user) == 300)  # значение из scenarios.behaviors.behavior_description


import requests
from timeout_decorator import timeout_decorator
from core.request.base_request import BaseRequest
from core.utils.exception_handlers import exc_handler
from core.monitoring.monitoring import monitoring


class RestNoMethodSpecifiedException(Exception):
    pass


class RestRequest(BaseRequest):
    on_exc_return = None
    BLOCKING_TIMEOUT = 0.2
    URL = "url"
    POST = "post"
    GET = "get"

    def __init__(self, items, id=None):
        super(RestRequest, self).__init__(items)
        self.url = items.get(self.URL)
        self.method = items.get("method", self.POST)
        self.rest_args = items.get("rest_args") or dict()

    @property
    def group_key(self):
        return self.URL

    @exc_handler(on_error_obj_method_name="on_timeout_error", handled_exceptions=(timeout_decorator.TimeoutError,))
    def run(self, data, params=None):
        if self.check_enabled():
            method = None
            if self.method == self.GET:
                method = self.get
            elif self.method == self.POST:
                method = self.post
            return method(data)

    def on_timeout_error(self, *args, **kwarg):
        monitoring.got_counter("core_rest_run_timeout")

    def _requests_get(self, params):
        return requests.get(self.url, params=params, **self.rest_args).text

    def _requests_post(self, data):
        return requests.post(self.url, data=data, **self.rest_args).text

    def get(self, data=None):
        return self._timeout_wrap(self._requests_get)(data)

    def post(self, data=None):
        return self._timeout_wrap(self._requests_post)(data)

    def __str__(self):
        return f"RestRequest: method={self.method} url={self.url} request_args={self.rest_args}"



import copy
from unittest import TestCase

from core.utils.masking_message import masking


class MaskingTest(TestCase):
    def test_bank_card(self):
        # 'message' in CARD_MASKING_FIELDS
        input_msg = {"message": "Слово до 1234567890123456"}
        expected = {"message": "Слово до ************3456"}
        masked_message = masking(input_msg)
        self.assertEqual(expected, masked_message)

        input_msg = {"message": "Слово до 1234567890123456 и после"}
        expected = {"message": "Слово до ************3456 и после"}
        masked_message = masking(input_msg)
        self.assertEqual(expected, masked_message)

        input_msg = {"message": "Склеено1234567890123456"}
        expected = {"message": "Склеено************3456"}
        masked_message = masking(input_msg)
        self.assertEqual(expected, masked_message)

        # если это поле входит в банковские, но не проходит по регулярке - то не маскируем
        input_msg = {"message": "1234"}
        expected = {"message": "1234"}
        masked_message = masking(input_msg)
        self.assertEqual(expected, masked_message)

        # маскировка так же применится к банковскому полю внутри коллекции
        input_msg = {'message': {'card': '1234567890123456'}}
        expected = {'message': {'card': '************3456'}}
        masked_message = masking(input_msg, masking_fields=['token'])
        self.assertEqual(expected, masked_message)

        # если это не баковское поле - то не маскируем
        input_msg = {'here_no_cards': {'no_card': '1234567890123456'}}
        expected = {'here_no_cards': {'no_card': '1234567890123456'}}
        masked_message = masking(input_msg)
        self.assertEqual(expected, masked_message)

        # проверки на целочисленный тип значений
        input_msg = {'message': 1234567890123456}
        expected = {'message': '************3456'}
        masked_message = masking(input_msg)
        self.assertEqual(expected, masked_message)

        input_msg = {'message': {'card': 1234567890123456}}
        expected = {'message': {'card': '************3456'}}
        masked_message = masking(input_msg)
        self.assertEqual(expected, masked_message)

        input_msg = {'message': 1234}
        expected = {'message': 1234}
        masked_message = masking(input_msg)
        self.assertEqual(expected, masked_message)

    def test_masking(self):
        masking_fields = {'spec_token': 2}
        input_msg = {"spec_token": '123456'}
        expected = {"spec_token": '***'}
        masked_message = masking(input_msg, masking_fields)
        self.assertEqual(expected, masked_message)

        # все простые типы маскируются как '***'
        input_msg = {"spec_token": {'int': 123, 'str': 'str', 'bool': True}}
        expected = {"spec_token": {'int': '***', 'str': '***', 'bool': '***'}}
        masked_message = masking(input_msg, masking_fields)
        self.assertEqual(expected, masked_message)

        # если маскируемое поле окажется внутри банковского поля - то оно маскируется с заданной вложеностью
        input_msg = {'message': {'spec_token': ['12', ['12', {'data': {'key': '12'}}]]}}
        expected = {'message': {'spec_token': ['***', ['***', '*items-1*collections-1*maxdepth-2*']]}}
        masked_message = masking(input_msg, masking_fields)
        self.assertEqual(expected, masked_message)

    def test_depth(self):
        # вложенность любой длины не маскируется пока не встретим ключ для маскировки
        masking_fields = ['spec_token']
        depth_level = 0
        input_msg = {'a': {'b': {'c': 1, 'spec_token': '123456'}}}
        expected = {'a': {'b': {'c': 1, 'spec_token': '***'}}}
        masked_message = masking(input_msg, masking_fields, depth_level)
        self.assertEqual(expected, masked_message)

        # проверка вложенной маскировки
        input_ = {'spec_token': [12, 12, {'key': [12, 12]}]}

        depth_level = 3
        expected = {'spec_token': ['***', '***', {'key': ['***', '***']}]}
        input_msg = copy.deepcopy(input_)
        masked_message = masking(input_msg, masking_fields, depth_level)
        self.assertEqual(expected, masked_message)

        depth_level = 2
        expected = {'spec_token': ['***', '***', {'key': '*items-2*collections-0*maxdepth-1*'}]}
        input_msg = copy.deepcopy(input_)
        masked_message = masking(input_msg, masking_fields, depth_level)
        self.assertEqual(expected, masked_message)

        depth_level = 1
        expected = {'spec_token': ['***', '***', '*items-2*collections-1*maxdepth-2*']}
        input_msg = copy.deepcopy(input_)
        masked_message = masking(input_msg, masking_fields, depth_level)
        self.assertEqual(expected, masked_message)

        depth_level = 0
        expected = {'spec_token': '*items-4*collections-2*maxdepth-3*'}
        input_msg = copy.deepcopy(input_)
        masked_message = masking(input_msg, masking_fields, depth_level)
        self.assertEqual(expected, masked_message)

    def test_check_set_components(self):
        input_message = {"set": {12, 13, 14}}
        expected = {"set": [12, 13, 14]}
        masked_message = masking(input_message)
        self.assertEqual(expected, masked_message)

        input_message = {"token": {12, 13, 14}}
        expected = {"token": "*items-3*collections-0*maxdepth-1*"}
        masked_message = masking(input_message)
        self.assertEqual(expected, masked_message)

        input_message = {"masked": {12, 13, 14}}
        expected = {"masked": ["***","***","***"]}
        masked_message = masking(input_message, masking_fields={"masked":2})
        self.assertEqual(expected, masked_message)



# coding: utf-8
import unittest

from core.names import field


class FieldTest1(unittest.TestCase):
    def test_fields(self):
        self.assertIsNotNone(field.PRONOUNCE_TEXT)
        self.assertIsNotNone(field.ITEMS)
        self.assertIsNotNone(field.MESSAGE_NAME)
        self.assertIsNotNone(field.AUTO_LISTENING)
        self.assertIsNotNone(field.FINISHED)
        self.assertIsNotNone(field.DEVICE)
        self.assertIsNotNone(field.SUB)
        self.assertIsNotNone(field.USER_ID)
        self.assertIsNotNone(field.USER_CHANNEL)
        self.assertIsNotNone(field.SMART_BIO)
        self.assertIsNotNone(field.SERVER_ACTION)
        self.assertIsNotNone(field.PROJECT_NAME)
        self.assertIsNotNone(field.INTENT)
        self.assertIsNotNone(field.INTENT_META)
        self.assertIsNotNone(field.APP_INFO)
        self.assertIsNotNone(field.PROJECT_ID)
        self.assertIsNotNone(field.APPLICATION_ID)
        self.assertIsNotNone(field.APP_VERSION_ID)
        self.assertIsNotNone(field.FRONTEND_ENDPOINT)
        self.assertIsNotNone(field.FRONTEND_TYPE)



# coding: utf-8
import json
import unittest
from unittest.mock import Mock

from smart_kit.message.as_is_to_message import AsIsToMessage


class TestAsIsTooMessage(unittest.TestCase):
    def setUp(self):
        self.command_ = Mock()
        self.request_ = Mock()
        self.message_ = Mock()
        self.command_.loader = "json.dumps"
        self.request_.header = "json"
        self.command_.payload = {}
        self.command_.root_nodes = {}
        self.command_.raw = {
            "content": {"test_param": ""},
            "surface": "some_surface",
            "project_id": "project_id",
        }
        self.message_.sub = 'sub'

    def test_as_is_to_message_as_dict(self):
        obj = AsIsToMessage(self.command_, self.message_, self.request_)
        self.assertEqual(obj.as_dict, {
            "content": {"test_param": ""},
            "surface": "some_surface",
            "project_id": "project_id"})



# -*- coding: utf-8 -*-
import unittest

from prometheus_client import Counter, Histogram

from core.monitoring.monitoring import Monitoring
from smart_kit.utils.picklable_mock import PicklableMock


class MonitoringTest(unittest.TestCase):
    def setUp(self):
        self.logger = PicklableMock()
        self.logger.exception = PicklableMock()
        self.config = PicklableMock()
        self.mock_rep = PicklableMock()
        self.monitoring = Monitoring()
        self.monitoring.apply_config({})

    def test_got_message_disabled(self):
        self.monitoring.turn_off()
        event_name = "test_counter"
        counter_item = self.monitoring._monitoring_items[self.monitoring.COUNTER]
        self.assertTrue(counter_item == dict())
        self.monitoring.got_counter(event_name)
        self.assertTrue(event_name not in counter_item)
        self.assertEqual(counter_item.get(event_name), None)

    def test_got_message_enabled(self):
        self.monitoring.turn_on()
        event_name = "test_counter"
        counter_item = self.monitoring._monitoring_items[self.monitoring.COUNTER]
        self.assertTrue(counter_item == dict())
        self.monitoring.got_counter(event_name)
        self.assertTrue(event_name in counter_item)
        self.assertEqual(type(counter_item[event_name]), type(Counter('counter_name', 'counter_name')))

    def test_got_histogram_disabled(self):
        self.monitoring.turn_off()
        event_name = "test_histogram"
        histogram_item = self.monitoring._monitoring_items[self.monitoring.HISTOGRAM]
        self.assertTrue(histogram_item == dict())
        histogram = self.monitoring._got_histogram(event_name)
        self.assertTrue(event_name not in histogram_item)
        self.assertIsNone(histogram)

    def test_got_histogram_enabled(self):
        self.monitoring.turn_on()
        event_name = "test_histogram"
        histogram_item = self.monitoring._monitoring_items[self.monitoring.HISTOGRAM]
        self.assertTrue(histogram_item == dict())
        self.monitoring._got_histogram(event_name)
        histogram_item = self.monitoring._monitoring_items[self.monitoring.HISTOGRAM]
        self.assertTrue(event_name in histogram_item)
        self.assertEqual(type(histogram_item[event_name]), type(Histogram('histogram_name', 'histogram_name')))

    def test_got_histogram_disabled_by_name(self):
        self.monitoring.turn_on()
        self.monitoring.disabled_metrics.append('test_.*')
        self.monitoring.disabled_metrics.append('.*_all')
        for event_name, disabled in [("test_one", True), ("test_two", True),
                                     ("not_a_test", False), ('metric_all', True)]:
            counter_item = self.monitoring._monitoring_items[self.monitoring.COUNTER]
            self.assertTrue(isinstance(counter_item, dict), event_name)
            self.monitoring.got_counter(event_name)
            if disabled:
                self.assertTrue(event_name not in counter_item, event_name)
            else:
                self.assertTrue(event_name in counter_item, event_name)

    def test_monitoring_init(self):
        from core.monitoring.monitoring import monitoring

        class MyCustomMonitoring(Monitoring):
            pass

        monitoring.set_instance(MyCustomMonitoring)
        self.assertEqual(type(monitoring.instance), MyCustomMonitoring)

        monitoring.set_instance(Monitoring)
        self.assertEqual(type(monitoring.instance), Monitoring)

    def test_decorator(self):
        from core.monitoring.monitoring import monitoring

        class MyCustomMonitoring(Monitoring):
            def got_histogram(self, name, description=None):
                def decorator(function):
                    def wrapper(*args, **kwargs):
                        result = function(*args, **kwargs) + " " + name
                        return result

                    return wrapper

                return decorator

        class SomeClass:
            @monitoring.got_histogram("test_histogram")
            def some_method(self):
                return "test"

        obj = SomeClass()
        self.assertEqual(obj.some_method(), "test")

        monitoring.set_instance(MyCustomMonitoring)
        self.assertEqual(obj.some_method(), "test test_histogram")

        monitoring.set_instance(Monitoring)
        self.assertEqual(obj.some_method(), "test")

        self.assertIn('test_histogram', monitoring._monitoring_items[Monitoring.HISTOGRAM].keys())




# coding: utf-8

from core.basic_models.actions.basic_actions import Action
from core.model.factory import factory, list_factory
from scenarios.scenario_models.field_requirements.field_requirements import FieldRequirement


class FieldValidator:
    def __init__(self, items):
        self._requirement = items.get("requirement")
        self._actions = items.get("actions")

        self.requirement = self.build_requirement()
        self.actions = self.build_actions()

    @factory(FieldRequirement)
    def build_requirement(self):
        return self._requirement

    @list_factory(Action)
    def build_actions(self):
        return self._actions



# coding: utf-8
from core.basic_models.scenarios.base_scenario import scenario_factory
from core.descriptions.smart_updatable_descriptions_items import SmartUpdatableDescriptionsItems


class ScenariosDescriptions(SmartUpdatableDescriptionsItems):
    def __init__(self, items):
        super(ScenariosDescriptions, self).__init__(scenario_factory, items, ordered=True)



import importlib
import os

ENVIRONMENT_VARIABLE = "SMART_KIT_APP_CONFIG"


def set_default(module, name, value):
    if hasattr(module, name):
        return

    setattr(module, name, value)


def get_static_path(app_config_path):
    project_folder, _ = os.path.split(app_config_path)
    static_path = os.path.join(project_folder, "./static")
    return static_path


def get_app_config(environment_variable=ENVIRONMENT_VARIABLE):
    app_config = os.getenv(environment_variable)
    app_config = importlib.import_module(app_config)

    static_path = get_static_path(app_config.__file__)
    set_default(app_config, "STATIC_PATH", static_path)
    set_default(app_config, "CONFIGS_PATH", os.path.join(static_path, "./configs"))
    set_default(app_config, "SECRET_PATH", os.path.join(static_path, "./configs"))
    references_path = os.path.join(static_path, "./references")
    set_default(app_config, "REFERENCES_PATH", references_path)

    # import and init monitoring first - because other classes use a singleton instance of Monitoring
    from core.monitoring.monitoring import Monitoring, monitoring
    set_default(app_config, 'MONITORING', Monitoring)
    monitoring.set_instance(app_config.MONITORING)

    from core.logging.logger_utils import LoggerMessageCreator
    from core.message.from_message import SmartAppFromMessage
    from scenarios.user.parametrizer import Parametrizer
    from scenarios.user.user_model import User
    from smart_kit.configs.settings import Settings
    from smart_kit.models.dialogue_manager import DialogueManager
    from smart_kit.models.smartapp_model import SmartAppModel
    from smart_kit.resources import SmartAppResources
    from smart_kit.start_points.main_loop_http import HttpMainLoop
    from smart_kit.start_points.postprocess import PostprocessMainLoop
    from smart_kit.testing.local import CLInterface
    from smart_kit.text_preprocessing.local_text_normalizer import LocalTextNormalizer
    from smart_kit.utils.cache import JSONCache
    from smart_kit.testing.suite import TestCase

    set_default(app_config, "LOCAL_TESTING", CLInterface)
    set_default(app_config, "TEST_CASE", TestCase)
    set_default(app_config, "NORMALIZER_ADDRESS", "http://127.0.0.1:9000")
    set_default(app_config, "PPS_URL", "")
    set_default(app_config, "NORMALIZER", LocalTextNormalizer())
    set_default(app_config, "USER", User)
    set_default(app_config, "LOGGER_MESSAGE_CREATOR", LoggerMessageCreator)
    set_default(app_config, "MAIN_LOOP", HttpMainLoop)
    set_default(app_config, "POSTPROCESSOR_MAIN_LOOP", PostprocessMainLoop)
    set_default(app_config, "PARAMETRIZER", Parametrizer)
    set_default(app_config, "MODEL", SmartAppModel)
    set_default(app_config, "DIALOGUE_MANAGER", DialogueManager)
    set_default(app_config, "SETTINGS", Settings)
    set_default(app_config, "RESOURCES", SmartAppResources)
    set_default(app_config, "FROM_MSG", SmartAppFromMessage)

    set_default(app_config, "NORMALIZATION_CACHE_TTL", 0)
    set_default(app_config, "NORMALIZATION_CACHE", JSONCache)

    set_default(app_config, "PLUGINS", ())
    set_default(app_config, "TO_MSG_VALIDATORS", ())
    set_default(app_config, "FROM_MSG_VALIDATORS", ())
    set_default(app_config, "AUTO_LISTENING", True)

    set_default(app_config, "STATIC_CLASSIFIERS_PATH", os.path.join(references_path, "./classifiers"))
    set_default(app_config, "STATIC_CLASSIFIERS_DATA_PATH", os.path.join(references_path, "./classifiers_data"))

    # Переменной можно присвоить значение среды, где запускается апп,
    # например: "ift", "uat", "pt", "prod" (это ИФТ, ПСИ, НТ, ПРОМ)
    set_default(app_config, "ENVIRONMENT", None)

    return app_config



import yaml
import os

from core.configs.base_config import BaseConfig
from core.db_adapter.ceph.ceph_adapter import CephAdapter
from core.db_adapter.os_adapter import OSAdapter
from core.repositories.file_repository import UpdatableFileRepository, FileRepository
from core.utils.singleton import Singleton


class Settings(BaseConfig, metaclass=Singleton):
    CephAdapterKey = "ceph"
    OSAdapterKey = "os"

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.configs_path = kwargs.get("config_path")
        self.references_path = kwargs.get("references_path")
        self.secret_path = kwargs.get("secret_path")
        self.app_name = kwargs.get("app_name")
        self.adapters = {Settings.CephAdapterKey: CephAdapter, self.OSAdapterKey: OSAdapter}
        self.repositories = [
            UpdatableFileRepository(
                self.subfolder_path("template_config.yml"), loader=yaml.safe_load, key="template_settings"
            ),
            FileRepository(self.subfolder_secret_path("kafka_config.yml"), loader=yaml.safe_load, key="kafka"),
            FileRepository(
                self.subfolder_path("ceph_config.yml"), loader=yaml.safe_load, key=self.CephAdapterKey
            ),
            FileRepository(self.subfolder_path("aiohttp.yml"), loader=yaml.safe_load, key="aiohttp"),
        ]
        self.repositories = self.override_repositories(self.repositories)
        self.init()

    def init(self):
        super().init()
        update_time = self["template_settings"].get("config_update_cooldown")
        if update_time is not None:
            for repo in self.repositories:
                if isinstance(repo, UpdatableFileRepository):
                    repo.update_cooldown = update_time

    def override_repositories(self, repositories: list):
        """
        Метод предназначен для переопределения репозиториев в дочерних классах.
        :param repositories: Список репозиториев родителя
        :return: Переопределённый в наследниках список репозиториев
        """
        return repositories

    @property
    def _subfolder(self):
        return self.configs_path

    def subfolder_secret_path(self, filename):
        return os.path.join(self.secret_path, filename)

    def get_source(self):
        if hasattr(self, "_source"):
            return self._source

        adapter_key = self.registered_repositories["template_settings"].data.get(
            "data_adapter") or Settings.OSAdapterKey
        adapter_settings = self.registered_repositories[
            adapter_key].data if adapter_key != Settings.OSAdapterKey else None
        adapter = self.adapters[adapter_key](adapter_settings)
        adapter.connect()
        source = adapter.source

        self._source = source
        return source



import pickle


def pickle_deepcopy(obj):
    if isinstance(obj, dict):
        return {pickle_deepcopy(k): pickle_deepcopy(v) for k, v in obj.items()}
    elif isinstance(obj, set):
        return {pickle_deepcopy(subobj) for subobj in obj}
    elif isinstance(obj, list):
        return [pickle_deepcopy(subobj) for subobj in obj]
    elif isinstance(obj, tuple):
        return tuple(pickle_deepcopy(subobj) for subobj in obj)
    if hasattr(obj, "__call__"):
        return obj
    return pickle.loads(pickle.dumps(obj, -1))



# coding: utf-8
from typing import Union, Any, Dict, List, Optional

from core.basic_models.operators.comparators import MoreComparator, LessComparator, MoreOrEqualComparator, \
    LessOrEqualComparator, EqualComparator, NotEqualComparator, InComparator, Comparator
from core.model.factory import build_factory, list_factory
from core.model.registered import Registered

operators = Registered()

operator_factory = build_factory(operators)


class Operator:

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        pass

    def compare(self, value: Any) -> bool:
        raise NotImplementedError


class CompositeOperator(Operator):
    operators: List[Operator]

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        super(CompositeOperator, self).__init__(items)
        self._operators: Dict[str, Any] = items["operators"]
        self.operators = self.build_operators()

    @list_factory(Operator)
    def build_operators(self):
        return self._operators

    def compare(self, value: Any) -> bool:
        return all(operator.compare(value) for operator in self.operators)


class AnyOperator(CompositeOperator):
    def compare(self, value: Any) -> bool:
        return any(operator.compare(value) for operator in self.operators)


class AmountOperator(Operator):
    amount: Union[int, str, List]
    comparator: Comparator

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        super(AmountOperator, self).__init__(items)
        self.amount = items["amount"]
        self.comparator: Optional[Comparator] = None

    def compare(self, value: Union[int, str]):
        return self.comparator.compare(value, self.amount)


class InOperator(AmountOperator):
    amount: Union[int, str, List]
    comparator: InComparator

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        super(InOperator, self).__init__(items)
        self.comparator = InComparator({})


class MoreOperator(AmountOperator):
    amount: Union[int, str, List]
    comparator: MoreComparator

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        super(MoreOperator, self).__init__(items)
        self.comparator = MoreComparator({})


class LessOperator(AmountOperator):
    amount: Union[int, str, List]
    comparator: LessComparator

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        super(LessOperator, self).__init__(items)
        self.comparator = LessComparator({})


class MoreOrEqualOperator(AmountOperator):
    amount: Union[int, str, List]
    comparator: MoreOrEqualComparator

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        super(MoreOrEqualOperator, self).__init__(items)
        self.comparator = MoreOrEqualComparator({})


class LessOrEqualOperator(AmountOperator):
    amount: Union[int, str, List]
    comparator: LessOrEqualComparator

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        super(LessOrEqualOperator, self).__init__(items)
        self.comparator = LessOrEqualComparator({})


class EqualOperator(AmountOperator):
    amount: Union[int, str, List]
    comparator: EqualComparator

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        super(EqualOperator, self).__init__(items)
        self.comparator = EqualComparator({})


class NotEqualOperator(AmountOperator):
    amount: Union[int, str, List]
    comparator: NotEqualComparator

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        super(NotEqualOperator, self).__init__(items)
        self.comparator = NotEqualComparator({})


class Exists(Operator):
    def compare(self, value: Any) -> bool:
        return True if value is not None else False


class EndsWithOperator(AmountOperator):
    """
    Оператор проверяет, оканчивается ли значение на amount
    """

    def compare(self, value: Union[int, str]) -> bool:
        return str(value).endswith(str(self.amount)) if value else False


class StartsWithOperator(AmountOperator):
    """
    Оператор проверяет, начинается ли значение с amount
    """

    def compare(self, value: Union[int, str]) -> bool:
        return str(value).startswith(str(self.amount)) if value else False



import copy

import aioredis
import typing

from aioredis.sentinel import Sentinel
from core.db_adapter.db_adapter import DBAdapter
from core.db_adapter import error
from core.monitoring.monitoring import monitoring
from core.logging.logger_utils import log


class AIORedisSentinelAdapter(DBAdapter):
    IS_ASYNC = True

    def __init__(self, config=None):
        super().__init__(config)
        self._sentinel: typing.Optional[Sentinel] = None
        self.service_name = None
        self.socket_timeout = None

        try:
            del self.config["type"]
        except KeyError:
            pass

    @monitoring.got_histogram("save_time")
    async def save(self, id, data):
        return await self._run(self._save, id, data)


    @monitoring.got_histogram("save_time")
    async def replace_if_equals(self, id, sample, data):
        return await self._run(self._replace_if_equals, id, sample, data)

    @monitoring.got_histogram("get_time")
    async def get(self, id):
        return await self._run(self._get, id)

    async def path_exists(self, path):
        return await self._run(self._path_exists, path)

    async def connect(self):

        config = copy.deepcopy(self.config)
        print("Here is the content of REDIS_CONFIG:", config)
        if not isinstance(config, dict):
            raise ValueError("REDIS_CONFIG should be a mapping")
        sentinels = config.pop("sentinels", None)
        self.service_name = config.pop("service_name", None)
        self.socket_timeout = config.get("socket_timeout", None)
        if not isinstance(sentinels, list):
            raise ValueError(
                "sentinels should be specified like [['sentinel.host1', 26379], ['sentinel.host2', 26379]]")
        sentinels_tuples = []
        for sent in sentinels:
            sentinels_tuples.append(tuple(sent))
        self._sentinel = Sentinel(sentinels_tuples, **config)

    def _open(self, filename, *args, **kwargs):
        pass

    async def _save(self, id, data):
        redis = await self._sentinel.master_for(self.service_name, socket_timeout=self.socket_timeout)
        await redis.set(id, data)

    async def _replace_if_equals(self, id, sample, data):
        log("no replace_if_equals method. running save instead", level="WARNING")
        await self.save(id, data)

    async def _get(self, id):
        redis = await self._sentinel.master_for(self.service_name, socket_timeout=self.socket_timeout)
        data = await redis.get(id)
        return data

    def _list_dir(self, path):
        raise error.NotSupportedOperation

    def _glob(self, path, pattern):
        raise error.NotSupportedOperation

    async def _path_exists(self, path):
        redis = await self._sentinel.master_for(self.service_name, socket_timeout=self.socket_timeout)
        return await redis.exists(path)

    def _on_prepare(self):
        pass



from unittest import TestCase

from core.utils.utils import time_check, current_time_ms


class TestTimeCheck(TestCase):

    def test_true(self):
        time = current_time_ms()
        self.assertTrue(time_check(time, 1 * 1000))

    def test_false(self):
        time = current_time_ms() - 40 * 1000
        self.assertFalse(time_check(time, 1 * 1000))



# coding: utf-8
import math
import socket
from collections import namedtuple
from time import time
from typing import Dict, List

import scenarios.logging.logger_constants as log_const
from core.basic_models.actions.command import Command
from core.logging.logger_utils import log
from core.names.field import APP_INFO
from core.text_preprocessing.preprocessing_result import TextPreprocessingResult
from core.utils.pickle_copy import pickle_deepcopy
from scenarios.actions.action_params_names import TO_MESSAGE_NAME, TO_MESSAGE_PARAMS, LOCAL_VARS
from core.monitoring.monitoring import monitoring


Callback = namedtuple(
            'Callback',
            'behavior_id expire_time scenario_id text_preprocessing_result action_params hostname')


class Behaviors:
    EXPIRATION_DELAY = 1

    def __init__(self, items, descriptions, user):
        self._items = items or {}
        self.descriptions = descriptions
        self._user = user
        self.Callback = Callback
        self._callbacks: Dict[str, Callback] = {}
        self._behavior_timeouts = []
        self._returned_callbacks = []

    def initialize(self):
        """Should be invoked after __init__ """
        for key, callback in self._items.items():
            callback.setdefault("text_preprocessing_result", {})
            callback.setdefault("action_params", {})
            callback.setdefault("hostname", None)
            self._callbacks[key] = self.Callback(**callback)

        callback_id = self._user.message.callback_id
        callback = self._get_callback(callback_id)
        if callback:
            callback_action_params = callback.action_params
            for key, value in callback_action_params.get(LOCAL_VARS, {}).items():
                self._user.local_vars.set(key, value)

    def _add_behavior_timeout(self, expire_time_us, callback_id):
        self._behavior_timeouts.append((expire_time_us, callback_id))

    def get_behavior_timeouts(self):
        return self._behavior_timeouts

    def _add_returned_callback(self, callback_id):
        self._returned_callbacks.append(callback_id)

    def get_returned_callbacks(self):
        return self._returned_callbacks

    def add(self, callback_id: str, behavior_id, scenario_id=None, text_preprocessing_result_raw=None,
            action_params=None):
        host = socket.gethostname()
        text_preprocessing_result_raw = text_preprocessing_result_raw or {}
        # behavior will be removed after now + timeout + EXPIRATION_DELAY
        expiration_time = (
                math.ceil(time()) +
                self.descriptions[behavior_id].timeout(self._user) +
                self.EXPIRATION_DELAY
        )
        action_params = action_params or dict()
        action_params[LOCAL_VARS] = pickle_deepcopy(self._user.local_vars.values)

        callback = self.Callback(
            behavior_id=behavior_id,
            expire_time=expiration_time,
            scenario_id=scenario_id,
            text_preprocessing_result=text_preprocessing_result_raw,
            action_params=action_params,
            hostname=host
        )
        self._callbacks[callback_id] = callback
        log(
            f"behavior.add: adding behavior %({log_const.BEHAVIOR_ID_VALUE})s with scenario_id"
            f" %({log_const.CHOSEN_SCENARIO_VALUE})s for callback %({log_const.BEHAVIOR_CALLBACK_ID_VALUE})s"
            f" expiration_time: %(expiration_time)s.",
            user=self._user,
            params={log_const.KEY_NAME: log_const.BEHAVIOR_ADD_VALUE,
                    log_const.BEHAVIOR_CALLBACK_ID_VALUE: callback_id,
                    log_const.BEHAVIOR_ID_VALUE: behavior_id,
                    log_const.CHOSEN_SCENARIO_VALUE: scenario_id,
                    "expiration_time": int(expiration_time)})

        behavior_description = self.descriptions[behavior_id]

        # add timer with now + behavior.timeout eval time
        expire_time_us = behavior_description.get_expire_time_from_now(self._user)
        self._add_behavior_timeout(expire_time_us, callback_id)

    def _delete(self, callback_id):
        if callback_id in self._callbacks:
            del self._callbacks[callback_id]

    def clear_all(self):
        self._callbacks = {}

    def _log_callback(self, callback_id: str, log_name: str, metric, behavior_result: str,
                      callback_action_params: Dict):
        callback = self._get_callback(callback_id)
        behavior = self.descriptions[callback.behavior_id] if callback else None
        callback_action_params = callback_action_params or {}
        to_message_name = callback_action_params.get(TO_MESSAGE_NAME, "UNKNOWN_MESSAGE_NAME")
        to_message_params = callback_action_params.get(TO_MESSAGE_PARAMS, {})
        app_info = to_message_params.get(APP_INFO, {})
        metric(self._user.settings.app_name, to_message_name)
        log_params = {log_const.KEY_NAME: log_name,
                      log_const.BEHAVIOR_CALLBACK_ID_VALUE: callback_id,
                      log_const.BEHAVIOR_ID_VALUE: str(behavior.id),
                      log_const.CHOSEN_SCENARIO_VALUE: callback.scenario_id,
                      "to_message_name": to_message_name,
                      "behavior_result": behavior_result}

        log(
            f"{self.__class__.__name__}.{behavior_result}: found valid behavior %({log_const.BEHAVIOR_ID_VALUE})s with scenario_id %({log_const.CHOSEN_SCENARIO_VALUE})s for callback %({log_const.BEHAVIOR_CALLBACK_ID_VALUE})s with to_message_name %(to_message_name)s",
            user=self._user,
            params=log_params)

    def success(self, callback_id: str) -> List[Command]:
        callback = self._get_callback(callback_id)
        result = []
        if callback:
            self._check_hostname(callback_id, callback)
            self._add_returned_callback(callback_id)
            behavior = self.descriptions[callback.behavior_id]
            callback_action_params = callback.action_params
            self._log_callback(
                callback_id,
                "behavior_success",
                monitoring.counter_behavior_success,
                "success",
                callback_action_params,
            )
            text_preprocessing_result = TextPreprocessingResult(callback.text_preprocessing_result)
            result = behavior.success_action.run(self._user, text_preprocessing_result, callback_action_params)
        else:
            log(f"behavior.success not found valid callback for callback_id {callback_id}",
                self._user,
                params={log_const.KEY_NAME: log_const.BEHAVIOR_SUCCESS_VALUE,
                        log_const.BEHAVIOR_CALLBACK_ID_VALUE: callback_id})
        self._delete(callback_id)
        return result

    def fail(self, callback_id: str) -> List[Command]:
        callback = self._get_callback(callback_id)
        result = []
        if callback:
            self._check_hostname(callback_id, callback)
            self._add_returned_callback(callback_id)
            behavior = self.descriptions[callback.behavior_id]
            callback_action_params = callback.action_params
            self._log_callback(callback_id, "behavior_fail",
                               monitoring.counter_behavior_fail, "fail",
                               callback_action_params)
            text_preprocessing_result = TextPreprocessingResult(callback.text_preprocessing_result)
            result = behavior.fail_action.run(self._user, text_preprocessing_result, callback_action_params)
        else:
            log(f"behavior.fail not found valid callback for callback_id {callback_id}",
                self._user,
                params={log_const.KEY_NAME: log_const.BEHAVIOR_FAIL_VALUE,
                        log_const.BEHAVIOR_CALLBACK_ID_VALUE: callback_id})
        self._delete(callback_id)
        return result

    def timeout(self, callback_id: str) -> List[Command]:
        callback = self._get_callback(callback_id)
        result = []
        if callback:
            self._add_returned_callback(callback_id)
            behavior = self.descriptions[callback.behavior_id]
            callback_action_params = callback.action_params
            self._log_callback(callback_id, "behavior_timeout",
                               monitoring.counter_behavior_timeout, "timeout",
                               callback_action_params)
            text_preprocessing_result = TextPreprocessingResult(callback.text_preprocessing_result)
            result = behavior.timeout_action.run(self._user, text_preprocessing_result, callback_action_params)
        else:
            log(f"behavior.timeout not found valid callback for callback_id {callback_id}",
                self._user,
                params={log_const.KEY_NAME: log_const.BEHAVIOR_TIMEOUT_VALUE,
                        log_const.BEHAVIOR_CALLBACK_ID_VALUE: callback_id})
        self._delete(callback_id)
        return result

    def misstate(self, callback_id: str) -> List[Command]:
        callback = self._get_callback(callback_id)
        result = []
        if callback:
            self._check_hostname(callback_id, callback)
            self._add_returned_callback(callback_id)
            behavior = self.descriptions[callback.behavior_id]
            callback_action_params = callback.action_params
            self._log_callback(callback_id, "behavior_misstate",
                               monitoring.counter_behavior_misstate, "misstate",
                               callback_action_params)
            text_preprocessing_result = TextPreprocessingResult(callback.text_preprocessing_result)
            result = behavior.misstate_action.run(self._user, text_preprocessing_result, callback_action_params)
        else:
            log(f"behavior.misstate not found valid callback for callback_id {callback_id}",
                self._user,
                params={log_const.KEY_NAME: log_const.BEHAVIOR_MISSTATE_VALUE,
                        log_const.BEHAVIOR_CALLBACK_ID_VALUE: callback_id})
        self._delete(callback_id)
        return result

    def _get_callback(self, callback_id):
        callback = self._callbacks.get(callback_id)
        return callback

    def has_callback(self, callback_id):
        callback = self._get_callback(callback_id)
        return callback is not None

    def get_callback_action_params(self, callback_id):
        callback = self._get_callback(callback_id)
        if callback:
            return callback.action_params

    def check_misstate(self, callback_id: str):
        log(f"behavior.check_misstate started: got callback %({log_const.BEHAVIOR_CALLBACK_ID_VALUE})s.",
            self._user,
            params={log_const.KEY_NAME: log_const.BEHAVIOR_CHECK_MISSTATE_VALUE,
                    log_const.BEHAVIOR_CALLBACK_ID_VALUE: callback_id})
        callback = self._get_callback(callback_id)
        if callback:
            callback_scenario_id = callback.scenario_id
            if callback_scenario_id is not None:
                last_scenario_equal_callback_scenario = self._user.last_scenarios.last_scenario_name != callback_scenario_id
                if not last_scenario_equal_callback_scenario:
                    log(
                        f"behavior.check_misstate: GOT MISSTATE for callback %({log_const.BEHAVIOR_CALLBACK_ID_VALUE})s: user_scenario: %(user_scenario)s, callback_scenario: %(callback_scenario)s.",
                        self._user,
                        params={log_const.KEY_NAME: log_const.BEHAVIOR_CHECK_MISSTATE_VALUE,
                                log_const.BEHAVIOR_CALLBACK_ID_VALUE: callback_id,
                                "user_scenario": self._user.last_scenarios.last_scenario_name,
                                "callback_scenario": callback_scenario_id},
                        level="WARNING")
                return last_scenario_equal_callback_scenario

    def expire(self):
        callback_id_for_delete = []
        for callback_id, (
                behavior_id, expiration_time, *_) in self._callbacks.items():
            if expiration_time <= time():
                callback_id_for_delete.append(callback_id)
        for callback_id in callback_id_for_delete:
            callback_action_params = self.get_callback_action_params(callback_id)
            to_message_name = callback_action_params.get(TO_MESSAGE_NAME)
            app_info = callback_action_params.get(APP_INFO, {})
            monitoring.counter_behavior_expire(self._user.settings.app_name, to_message_name)
            log_params = {log_const.KEY_NAME: "behavior_expire",
                          log_const.BEHAVIOR_CALLBACK_ID_VALUE: callback_id,
                          log_const.BEHAVIOR_DATA_VALUE: str(self._callbacks[callback_id]),
                          "to_message_name": to_message_name}

            log(
                f"behavior.expire: if you see this - something went wrong(should be timeout in normal case) callback %({log_const.BEHAVIOR_CALLBACK_ID_VALUE})s,  with to_message_name %(to_message_name)s",
                params=log_params, level="WARNING", user=self._user)
            self._delete(callback_id)

    def check_got_saved_id(self, behavior_id):
        if self.descriptions[behavior_id].loop_def:
            for callback_id, (_behavior_id, *_) in self._callbacks.items():
                if _behavior_id == behavior_id:
                    log(
                        f"behavior.check_got_saved_id == True: already got saved behavior %({log_const.BEHAVIOR_ID_VALUE})s for callback_id %({log_const.BEHAVIOR_CALLBACK_ID_VALUE})s",
                        user=self._user,
                        params={log_const.KEY_NAME: "behavior_got_saved",
                                log_const.BEHAVIOR_CALLBACK_ID_VALUE: callback_id,
                                log_const.BEHAVIOR_ID_VALUE: behavior_id,
                                log_const.BEHAVIOR_DATA_VALUE: str(self._callbacks[callback_id])})
                    return True
            return False
        else:
            return False

    @property
    def raw(self):
        return {key: callback._asdict() for key, callback in self._callbacks.items()}

    def _get_to_message_name(self, callback_id):
        callback_action_params = self.get_callback_action_params(callback_id) or {}
        to_message_name = callback_action_params.get(TO_MESSAGE_NAME)
        return to_message_name

    def _check_hostname(self, callback_id, callback) -> None:
        host: str = socket.gethostname()
        if callback.hostname != host:
            log(
                f"behavior.check_hostname: current %({log_const.HOSTNAME})s and "
                f"last %({log_const.LAST_HOSTNAME})s hostname are not equal on the same message_id.",
                user=self._user,
                params={log_const.KEY_NAME: log_const.CHECK_HOSTNAME,
                        log_const.BEHAVIOR_CALLBACK_ID_VALUE: callback_id,
                        log_const.BEHAVIOR_ID_VALUE: callback.behavior_id,
                        log_const.CHOSEN_SCENARIO_VALUE: callback.scenario_id,
                        log_const.MESSAGE_ID: self._user.message.logging_uuid,
                        log_const.HOSTNAME: host,
                        log_const.LAST_HOSTNAME: callback.hostname,
                        }
            )
            monitoring.counter_host_has_changed(self._user.settings.app_name)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: SmartAppToMessage.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='SmartAppToMessage.proto',
  package='tutorial',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x17SmartAppToMessage.proto\x12\x08tutorial\x1a\x1cgoogle/protobuf/struct.proto\"\xae\x02\n\x11SmartAppToMessage\x12\x11\n\tmessageId\x18\x01 \x01(\x03\x12\x11\n\tsessionId\x18\x02 \x01(\t\x12\x13\n\x0bmessageName\x18\x03 \x01(\t\x12\x39\n\x07payload\x18\x04 \x03(\x0b\x32(.tutorial.SmartAppToMessage.PayloadEntry\x12.\n\x04uuid\x18\x05 \x01(\x0b\x32 .tutorial.SmartAppToMessage.UUID\x1a\x46\n\x0cPayloadEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12%\n\x05value\x18\x02 \x01(\x0b\x32\x16.google.protobuf.Value:\x02\x38\x01\x1a+\n\x04UUID\x12\x0e\n\x06userId\x18\x01 \x01(\t\x12\x13\n\x0buserChannel\x18\x02 \x01(\t'
  ,
  dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])




_SMARTAPPTOMESSAGE_PAYLOADENTRY = _descriptor.Descriptor(
  name='PayloadEntry',
  full_name='tutorial.SmartAppToMessage.PayloadEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='tutorial.SmartAppToMessage.PayloadEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='tutorial.SmartAppToMessage.PayloadEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=255,
  serialized_end=325,
)

_SMARTAPPTOMESSAGE_UUID = _descriptor.Descriptor(
  name='UUID',
  full_name='tutorial.SmartAppToMessage.UUID',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='userId', full_name='tutorial.SmartAppToMessage.UUID.userId', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='userChannel', full_name='tutorial.SmartAppToMessage.UUID.userChannel', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=327,
  serialized_end=370,
)

_SMARTAPPTOMESSAGE = _descriptor.Descriptor(
  name='SmartAppToMessage',
  full_name='tutorial.SmartAppToMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='messageId', full_name='tutorial.SmartAppToMessage.messageId', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sessionId', full_name='tutorial.SmartAppToMessage.sessionId', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='messageName', full_name='tutorial.SmartAppToMessage.messageName', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payload', full_name='tutorial.SmartAppToMessage.payload', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='uuid', full_name='tutorial.SmartAppToMessage.uuid', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_SMARTAPPTOMESSAGE_PAYLOADENTRY, _SMARTAPPTOMESSAGE_UUID, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=68,
  serialized_end=370,
)

_SMARTAPPTOMESSAGE_PAYLOADENTRY.fields_by_name['value'].message_type = google_dot_protobuf_dot_struct__pb2._VALUE
_SMARTAPPTOMESSAGE_PAYLOADENTRY.containing_type = _SMARTAPPTOMESSAGE
_SMARTAPPTOMESSAGE_UUID.containing_type = _SMARTAPPTOMESSAGE
_SMARTAPPTOMESSAGE.fields_by_name['payload'].message_type = _SMARTAPPTOMESSAGE_PAYLOADENTRY
_SMARTAPPTOMESSAGE.fields_by_name['uuid'].message_type = _SMARTAPPTOMESSAGE_UUID
DESCRIPTOR.message_types_by_name['SmartAppToMessage'] = _SMARTAPPTOMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SmartAppToMessage = _reflection.GeneratedProtocolMessageType('SmartAppToMessage', (_message.Message,), {

  'PayloadEntry' : _reflection.GeneratedProtocolMessageType('PayloadEntry', (_message.Message,), {
    'DESCRIPTOR' : _SMARTAPPTOMESSAGE_PAYLOADENTRY,
    '__module__' : 'SmartAppToMessage_pb2'
    # @@protoc_insertion_point(class_scope:tutorial.SmartAppToMessage.PayloadEntry)
    })
  ,

  'UUID' : _reflection.GeneratedProtocolMessageType('UUID', (_message.Message,), {
    'DESCRIPTOR' : _SMARTAPPTOMESSAGE_UUID,
    '__module__' : 'SmartAppToMessage_pb2'
    # @@protoc_insertion_point(class_scope:tutorial.SmartAppToMessage.UUID)
    })
  ,
  'DESCRIPTOR' : _SMARTAPPTOMESSAGE,
  '__module__' : 'SmartAppToMessage_pb2'
  # @@protoc_insertion_point(class_scope:tutorial.SmartAppToMessage)
  })
_sym_db.RegisterMessage(SmartAppToMessage)
_sym_db.RegisterMessage(SmartAppToMessage.PayloadEntry)
_sym_db.RegisterMessage(SmartAppToMessage.UUID)


_SMARTAPPTOMESSAGE_PAYLOADENTRY._options = None
# @@protoc_insertion_point(module_scope)
